;; -*- Mode: Scheme; tab-width: 4 -*-

;; 64-bit version of bootstrap.scm

;; build an image file from scratch, given the code, and a vector of
;; constants.

(define code-type 'bytecode)

;; utilities
(define (shift-left val n)
  (if (zero? n)
      val
      (shift-left (* val 2) (- n 1))))

(define (shift-right val n)
  (if (zero? n)
      val
      (shift-right (/ val 2) (- n 1))))

(define tag-value
  (let ((tags
		 (map (lambda (x)
				(list (car x)
					  (shift-left (cadr x) (caddr x))))
			  ;; kind        value shift
			  '(;; immediate types
				(CHAR		1  1)
				(BOOL		3  1)
				(NIL		5  1)
				(UNDEFINED	7  1)
				;; pointer types
				(ENV		2  2)
				(CLOSURE	3  2)
				(SAVE		4  2)
				(STRING		5  2)
				(VECTOR		6  2)
				(TUPLE		7  2)
				(CONS		8  2)
				(SYMBOL		9  2)
				(CODE		10 2)
				))))
	(lambda (kind value)
	  (+ (cadr (assq kind tags)) (shift-left value 8)))))

(define (write-bootstrap-file output-port code-string constants)

  (let ((offset 0)
		(symbol-offsets '()))
	
	(define (big-encode-long n)
	  ;; encode an integer as eight big-endian bytes
	  (if (< n 0)
		  (set! n (+ #x10000000000000000 n)))
	  (let loop ((i 8)
				 (n n)
				 (r '()))
		(if (zero? i)
			r
			(loop (- i 1)
				  (quotient n #x100)
				  (cons (modulo n #x100) r)))))

	(define (emit byte)
	  (write-char (ascii->char byte) output-port)
	  (set! offset (+ offset 1)))

	(define (emit-long n)
	  (for-each emit (big-encode-long n)))

	(define (how-many x n)
	  (quotient (- (+ x n) 1) n))

	;; we return the offset of the object written. for example, when
	;; writing a pair, we write the car, write the cdr, then emit a
	;; cons cell containing the pointers to each, and finally return
	;; the offset of that cons cell.

	(define (emit-object x)
	  ;; immediates: char, bool, nil, undefined
	  ;; tuples: env, closure, save, string, vector, tuple, cons, symbol, code
	  (cond ((integer? x)
			 (+ 1 (shift-left x 1)))
			((char? x)
			 (tag-value 'CHAR (char->ascii x)))
			((boolean? x)
			 (tag-value 'BOOL (if x 1 0)))
			((null? x)
			 (tag-value 'NIL 0))
			((pair? x)
			 (let ((car-offset (emit-object (car x)))
				   (cdr-offset (emit-object (cdr x))))
			   (let ((result offset))
				 (emit-long (tag-value 'CONS 2))
				 (emit-long car-offset)
				 (emit-long cdr-offset)
				 result)))
			((vector? x)
			 (let ((values (map emit-object (vector->list x))))
			   (let ((result offset))
				 (emit-long (tag-value 'VECTOR (vector-length x)))
				 (for-each emit-long values)
				 result)))
			((string? x)
			 (let ((len (string-length x)))
			   (let ((wlen (how-many len 8))
					 (result offset))
				 (emit-long (tag-value 'STRING (+ wlen 1)))
				 ;; emit the string length
				 (emit-long (+ 1 (shift-left len 1)))
				 ;; emit the string
				 (for-each
				  (lambda (ch) (emit (char->ascii ch)))
				  (string->list x))
				 ;; pad the string
				 (let loop ((n (- (* wlen 8) len)))
				   (cond ((not (zero? n))
						  (emit 0)
						  (loop (- n 1)))))
				 result)))
			((symbol? x)
			 (let ((probe (assq x symbol-offsets)))
			   (if probe
				   (cdr probe)
				   (let ((string-offset (emit-object (symbol->string x))))
					 (let ((result offset))
					   (emit-long (tag-value 'SYMBOL 1))
					   (emit-long string-offset)
					   (set! symbol-offsets (cons (cons x result) symbol-offsets))
					   result)))))
			(else
			 (error "unsupported object type"))))

	(define (emit-symbol-vector)
	  (let ((result offset))
		(emit-long (tag-value 'VECTOR (length symbol-offsets)))
		(for-each
		 (lambda (entry) (emit-long (cdr entry)))
		 symbol-offsets)
		result))

	;; emit a header
	(for-each
	 (lambda (ch) (write-char ch output-port))
	 (string->list "(lumberjack-image boot-strap)"))
	(write-char (ascii->char 10) output-port)

	;; structure of a bootstrap image:
	;; <space for gc-relocate offset> <constants> ... <constant-vector> <code-object> <closure> <offset to closure>
	(emit-long 1)

	(let ((cv-offset (emit-object (list->vector constants)))
		  (cs-offset (emit-object code-string))
		  (sv-offset (emit-symbol-vector)))
	  ;; --------------------------------------------------
	  ;; emit the code object
	  ;; --------------------------------------------------
	  (let ((code-offset offset))
		(emit-long (tag-value 'CODE 2))
		;; pointer to constant vector
		(emit-long cv-offset)
		;; pointer to code string
		(emit-long cs-offset)
		(let ((closure-offset offset))
		  ;; closure = { tag, code, pc, lenv }
		  (emit-long (tag-value 'CLOSURE 3))
		  ;; pointer to code object
		  (emit-long code-offset)
		  ;; pc (remember, BOX_INTEGER(0) == 1)
		  (emit-long 1)
		  ;; lenv
		  (emit-long sv-offset)
		  ;; *** HACK ALERT ***
		  ;; The environment pointer is deliberately not set to zero due to confusion on
		  ;; the part of gc_relocate, which will think this points back to the beginning of
		  ;; the image/heap.  But we are also abusing the environment pointer as a way of
		  ;; passing in the constant-vector to the bootstrap code.  This can be used to
		  ;; initialize the runtime's symbol table.
		  ;; *** HACK ALERT ***

		  (emit-long closure-offset)
		  ))
	  )
	)
  )


(define (bootstrap code)
  (let ((tc (transform code (lambda (x) x) #t)))
	(let ((top-level-symbols (cadddr (car tc))))
	  (if (eq? (car top-level-symbols) 'let-macro)
		  (set! top-level-symbols (caddr (car (caddr top-level-symbols))));; bad sam
		  (set! top-level-symbols (cadr (car top-level-symbols))))
	  (display top-level-symbols)
	  (let ((cc (compile tc '())))
		(let ((vm-pair ((make-bytecoder) cc)))
		  (let ((cs (car vm-pair))
				(cv (cdr vm-pair)))
			(let ((op (open-output-file "test.image")))
			  (write-bootstrap-file op cs (append cv (list top-level-symbols)))
			  (close-output-port op)
			  )))))))
								  
(define (bootstrap-simple code)
  (let ((tc (transform code (lambda (x) x) #t)))
	(let ((cc (compile tc '())))
	  (let ((vm-pair ((make-bytecoder) cc)))
		(let ((cs (car vm-pair))
			  (cv (cdr vm-pair)))
		  (let ((op (open-output-file "test.image")))
			(write-bootstrap-file op cs (append cv (list)))
			(close-output-port op)
			))))))

(define (bootstrap-with-lunacy)
  (%print "reading lunacy.scm...")
  (let ((forms (car (read-forms-from-file "lunacy.scm"))))
	(%print "bootstrapping...")
	(bootstrap forms)))

(define (bootstrap-with-guile)
  (let ((forms (car (read-forms-from-file "lunacy.scm"))))
	(bootstrap forms)))

(define (wrap-file-with-let filename)
  (with-input-from-file filename
	(lambda ()
	  (let loop ((form (read))
				 (forms '()))
		(if (eof-object? form)
			`(let () ,@(reverse forms))
			(loop (read) (cons form forms)))))))

(define (wrap-files-with-let . files)
  (let loop1 ((all-forms '())
			  (files files))
	(if (null? files)
		`(let () ,@all-forms)
		(with-input-from-file (car files)
			(lambda ()
			  (let loop2 ((form (read))
						  (forms '()))
				(if (eof-object? form)
					(loop1
					 (append all-forms (reverse forms))
					 (cdr files))
					(loop2 (read) (cons form forms)))))))))
