// -*- Mode: C; tab-width: 4 -*-

#include <stdio.h>
#include "machine.h"

// assumes argument tuple is always named 'val'
// [it should be changed to 'args']
#define GET_ARG(n)	((SCHEME *)val[(n)+2])
#define GET_IARG(n)	(UNBOX_INTEGER((SCHEME)val[(n)+2]))

//
// ===========================================================================
//			  builtin primitives
// ===========================================================================
//

void print_list					(SCHEME * l);
void print_vector				(SCHEME * v);
void print_tuple				(SCHEME * v);
void print_tuple_contents		(SCHEME * v);
void print_tuple_addresses		(SCHEME * v);
void print_string				(SCHEME * s);
void print_symbol				(SCHEME * s);

// many print-related builtins can eventually be replaced by
// implementations written in scheme.  these are handy for bootstrapping
// and debugging.

SCHEME *
builtin_print_object (SCHEME * obj)
{
  if ((!obj) || (IMMEDIATE(obj))) {
    SCHEME imm = (SCHEME) obj;
    if (IS_INTEGER(obj)) {
      fprintf (stdout, "%d", UNBOX_INTEGER ((long)obj));
    } else {
      switch (GET_TYPECODE(imm)) {
      case TC_CHAR:
		fputs ("#\\", stdout);
		fputc (GET_PAYLOAD(imm), stdout);
		break;
      case TC_BOOL:
		fputs ((imm == LJ_TRUE ? "#t" : "#f"), stdout);
		break;
	  case TC_NIL:
		fputs ("()", stdout);
		break;
	  case TC_UNDEFINED:
		fputs ("#{undefined}", stdout);
		break;
      default:
		fprintf (stdout, "{TC:%02x PAYLOAD:%06x}", GET_TYPECODE(imm), GET_PAYLOAD(imm));
		break;
      }
    }
  } else {
    switch (GET_TYPECODE (*obj)) {
    case TC_PAIR:
      print_list (obj);
      break;
    case TC_VECTOR:
      print_vector (obj);
      break;
    case TC_TUPLE:
	  fprintf (stdout, "<tuple ");
	  print_tuple_addresses (obj);
	  fprintf (stdout, ">");
      break;
	case TC_STRING:
	  print_string (obj);
	  break;
	case TC_SYMBOL:
	  print_symbol (obj);
	  break;
	case TC_CODE:
	  fprintf (stdout, "<code ");
	  print_tuple_addresses (obj);
	  fprintf (stdout, ">");
	  break;
	case TC_CLOSURE:
	  fprintf (stdout, "<closure ");
	  print_tuple_addresses (obj);
	  fprintf (stdout, ">");
	  break;
	case TC_SAVE:
	  fprintf (stdout, "<save ");
	  print_tuple_addresses (obj);
	  fprintf (stdout, ">");
	  break;
    default:
      fprintf (stdout, "{TC:%02x LENGTH:%06x}", GET_TYPECODE(*obj), GET_TUPLE_LENGTH(*obj));
      break;
    }
  }
}

void
print_list (SCHEME * l)
{
  fprintf (stdout, "(");
  while (1) {
    SCHEME * car = (SCHEME *) l[1];
    SCHEME * cdr = (SCHEME *) l[2];

    builtin_print_object (car);
    if ((ulong)cdr == LJ_NIL) {
      fprintf (stdout, ")");
      break;
    } else if ((!IMMEDIATE(cdr)) && (GET_TYPECODE (*cdr) == TC_PAIR)) {
      fprintf (stdout, " ");
      l = cdr;
    } else {
      fprintf (stdout, " . ");
      builtin_print_object (cdr);
      fprintf (stdout, ")");
      break;
    }
  }
}

void print_vector (SCHEME * v)
{
  fprintf (stdout, "#(");
  print_tuple_contents (v);
  fprintf (stdout, ")");
}

void print_tuple (SCHEME * v)
{
  unsigned long l = GET_TUPLE_LENGTH(*v);
  unsigned long i;
  fprintf (stdout, "#{env.next:%08x ", v[1]);
  for (i=1; i < l; i++) {
    builtin_print_object ((SCHEME *) v[i+1]);
    if (i < (l-1)) {
      fprintf (stdout, " ");
	}
  }
  fprintf (stdout, "}");
}

void
print_tuple_contents (SCHEME *v)
{
  unsigned long l = GET_TUPLE_LENGTH (*v);
  unsigned long i;

  for (i=0; i < l; i++) {
    builtin_print_object ((SCHEME *) v[i+1]);
    if (i < (l-1)) {
      fprintf (stdout, " ");
	}
  }
}

void
print_tuple_addresses (SCHEME *v)
{
  unsigned long l = GET_TUPLE_LENGTH (*v);
  unsigned long i;

  for (i=0; i < l; i++) {
	fprintf (stdout, "%08x", v[i+1]);
    if (i < (l-1)) {
      fprintf (stdout, " ");
	}
  }
}


void
print_string (SCHEME * s)
{
  ulong length = (ulong) UNBOX_INTEGER(s[1]);
  ulong i;
  char * ps = (char *) &(s[2]);
  fputc ('"', stdout);
  for (i=0; i < length; i++, ps++) {
	if (*ps == '"') {
	  fputc ('\\', stdout);
	  fputc ('"', stdout);
	} else {
	  fputc (*ps, stdout);
	}
  }
  fputc ('"', stdout);
}

void
print_symbol (SCHEME *sym)
{
  SCHEME *s = (SCHEME *) sym[1];
  ulong length = (ulong) UNBOX_INTEGER(s[1]);
  ulong i;
  char * ps = (char *) &(s[2]);
  for (i=0; i < length; i++, ps++) {
	fputc (*ps, stdout);
  }
}

SCHEME *
builtin_print (SCHEME * val)
{
  builtin_print_object ((SCHEME *)val[2]);
  fputc ('\n', stdout);
  return (SCHEME *) LJ_UNDEFINED;
}

SCHEME *
builtin_read_char (SCHEME * val)
{
  return (SCHEME *) TAG_VALUE (TC_CHAR, fgetc(stdin));
}

SCHEME *
builtin_write_char (SCHEME * val)
{
  // XXX typecheck
  SCHEME * ch = ((SCHEME *)val[2]);
  fputc (GET_PAYLOAD(ch), stdout);
  return (SCHEME *) LJ_UNDEFINED;
}


SCHEME *
builtin_string_compare (SCHEME * val)
{
  // XXX typecheck
  SCHEME * s1 = ((SCHEME *)val[2]);
  SCHEME * s2 = ((SCHEME *)val[3]);
  SCHEME l1 = s1[1];
  SCHEME l2 = s2[1];
  
  if (l1 != l2) {
	return (SCHEME *) BOX_INTEGER (l1 < l2 ? -1 : +1);
  } else {
	return (SCHEME *) BOX_INTEGER (memcmp ((void*)&(s1[2]), (void*)&(s2[2]), UNBOX_INTEGER(l1)));
  }
}

SCHEME *
builtin_string_ref (SCHEME * val)
{
  SCHEME * s = ((SCHEME *)val[2]);
  SCHEME   i = UNBOX_INTEGER (val[3]);
  SCHEME  ls = UNBOX_INTEGER (s[1]);
  if (i < ls) {
	char * p = (char *)&(s[2]);
	return (SCHEME *) TAG_VALUE (TC_CHAR, p[i]);
  } else {
	return NULL;
  }
}

SCHEME *
builtin_string_set (SCHEME * val)
{
  // XXX typecheck
  SCHEME * s = ((SCHEME *)val[2]);
  SCHEME   i = UNBOX_INTEGER (val[3]);
  SCHEME  ch = GET_PAYLOAD (val[4]);
  SCHEME  ls = UNBOX_INTEGER (s[1]);
  if (i < ls) {
	char * p = (char *)&(s[2]);
	p[i] = ch;
	return (SCHEME *) LJ_UNDEFINED;
  } else {
	return NULL;
  }
}

// on some machines memmove can turn into
// inline assembly.  investigate.

SCHEME *
builtin_buffer_copy (SCHEME * val)
{
  // XXX typecheck
  // src src-start n dst dst-start
  SCHEME * src	= ((SCHEME *)val[2]);
  SCHEME ss		= UNBOX_INTEGER(val[3]);
  SCHEME n		= UNBOX_INTEGER(val[4]);
  SCHEME * dst	= ((SCHEME *)val[5]);
  SCHEME ds		= UNBOX_INTEGER(val[6]);
  char * psrc = ((char *)&(src[2])) + ss;
  char * pdst = ((char *)&(dst[2])) + ds;
  memmove ((void *) pdst, (void *) psrc, (size_t) n);
  return (SCHEME *) LJ_UNDEFINED;
}

SCHEME *
builtin_write_string (SCHEME * val)
{
  SCHEME * s = ((SCHEME *)val[2]);
  SCHEME   i = UNBOX_INTEGER (val[3]);
  SCHEME  ls = UNBOX_INTEGER (s[1]);
  void * p = (void *)&(s[2]);
  fwrite (p, 1, ls, stdout);
  return (SCHEME *) LJ_UNDEFINED;
}

SCHEME *
builtin_flush (SCHEME * val)
{
  fflush (stdout);
  return (SCHEME *) LJ_UNDEFINED;
}

SCHEME *
builtin_assq (SCHEME * val)
{
  SCHEME * sym = GET_ARG(0);
  SCHEME * alist = GET_ARG(1);
  
  while (alist != (SCHEME *) LJ_NIL) {
	SCHEME * cell = (SCHEME *) alist[1];
	if (sym == (SCHEME *) cell[1]) {
	  return cell;
	} else {
	  alist = (SCHEME *) alist[2];
	}
  }
  return (SCHEME *) LJ_FALSE;
}

SCHEME *
builtin_memq (SCHEME * val)
{
  SCHEME * x = GET_ARG(0);
  SCHEME * list = GET_ARG(1);
  
  while (list != (SCHEME *) LJ_NIL) {
	if (x == (SCHEME *) list[1]) {
	  return list;
	} else {
	  list = (SCHEME *) list[2];
	}
  }
  return (SCHEME *) LJ_FALSE;
}

// unix-specific-ish

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

SCHEME *
builtin_open (SCHEME * val)
{
  char * filename = GET_STRING_POINTER (GET_ARG(0));
  SCHEME * create_flag = GET_ARG(1);
  SCHEME * write_flag  = GET_ARG(2);
  int mode = GET_IARG(3);
  int flags = 0;

  if (create_flag != LJ_FALSE) {
	flags |= O_CREAT;
  }
  if (write_flag != LJ_FALSE) {
	flags |= O_RDWR;
  } else {
	flags |= O_RDONLY;
  }

  return (SCHEME *) BOX_INTEGER (open (filename, flags, mode));
}

SCHEME *
builtin_read (SCHEME * val)
{
  int fd		= GET_IARG(0);
  char * buf	= GET_STRING_POINTER (GET_ARG(1));
  int len		= GET_IARG(2);
  return (SCHEME *) BOX_INTEGER (read (fd, buf, len));
}

SCHEME *
builtin_write (SCHEME * val)
{
  int fd		= GET_IARG(0);
  char * buf	= GET_STRING_POINTER (GET_ARG(1));
  int len		= GET_IARG(2);
  return (SCHEME *) BOX_INTEGER (write (fd, buf, len));
}

SCHEME *
builtin_close (SCHEME * val)
{
  int fd = GET_IARG(0);
  return (SCHEME *) BOX_INTEGER (close (fd));
}

#include <time.h>
SCHEME *
builtin_clock (SCHEME * val)
{
  // overflow
  return (SCHEME *) BOX_INTEGER (clock());
}

// ===========================================================================
//                 rudimentary inet socket support
// ===========================================================================

// to do in socket stuff, we need:
// 1) socket
// 2) connect
// 3) listen
// 4) accept
// 5) recv
// 6) send
// 7) int
// close inet_aton (const char *NAME, struct in_addr *ADDR)

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

SCHEME *
builtin_socket (SCHEME *val)
{
  int s_family = GET_IARG(0);
  int s_type   = GET_IARG(1);
  return (SCHEME *) BOX_INTEGER (socket (s_family, s_type, 0)); // default protocol to zero
}

SCHEME *
// inet only for now
builtin_connect (SCHEME *val)
{
  int s			= GET_IARG(0);
  char * addr	= GET_STRING_POINTER(GET_ARG(1));
  int port		= GET_IARG(2);
  struct sockaddr_in a;
  
  if (inet_aton (addr, &(a.sin_addr)) != INADDR_NONE) {
	// converted the address, let's connect now
	a.sin_family = AF_INET;
	a.sin_port = htons ((short)port);
	return (SCHEME *)BOX_INTEGER (connect (s, (struct sockaddr *)&a, sizeof(struct sockaddr)));
  } else {
	return (SCHEME *)BOX_INTEGER (-1);
  }
}

SCHEME *
// inet only for now
builtin_bind (SCHEME *val)
{
  int s			= GET_IARG(0);
  int d0		= GET_IARG(1);
  int d1		= GET_IARG(2);
  int d2		= GET_IARG(3);
  int d3		= GET_IARG(4);
  int port		= GET_IARG(5);
  struct sockaddr_in a;
  int result;
  
  a.sin_family = AF_INET;
  a.sin_port = htons ((short)port);
  a.sin_addr.s_addr = htonl ((d0<<24) | (d1<<16) | (d2<<8) | (d3<<0));
  // a.sin_addr.s_addr = htonl (INADDR_ANY);
  fprintf (stderr, "%s:%d\n", inet_ntoa (a.sin_addr), port);
  result = bind (s, (struct sockaddr *)&a, sizeof(a));
  if (result == -1) {
	perror("bind");
  } else {
	return (SCHEME *)BOX_INTEGER (result);
  }
}

SCHEME *
builtin_listen (SCHEME *val)
{
  int s = GET_IARG(0);
  int n = GET_IARG(1);
  return (SCHEME *) BOX_INTEGER (listen (s, n));
}

SCHEME *
// inet only for now
builtin_accept (SCHEME *val)
{
  // XXX: return the address
  int s = GET_IARG(0);
  int addrlen;
  struct sockaddr_in a;
  return (SCHEME *) BOX_INTEGER (accept (s, (struct sockaddr *)&a, &addrlen));
}

#if 0
SCHEME *
builtin_strerror (SCHEME *val)
{
  char * e = strerror (errno);
  int el = strlen(e);
  // need to pass in a buffer...
}
#endif

builtin_function builtin_table[] = {
  builtin_print,
  builtin_read_char,
  builtin_write_char,
  builtin_string_compare,
  builtin_string_ref,
  builtin_string_set,
  builtin_buffer_copy,
  builtin_write_string,
  builtin_flush,
  builtin_assq,
  builtin_memq,
  builtin_open,
  builtin_read,
  builtin_write,
  builtin_close,
  builtin_socket,
  builtin_connect,
  builtin_bind,
  builtin_listen,
  builtin_accept,
  builtin_clock,
  NULL
};
