;; -*- Mode: Scheme; tab-width: 4 -*-

(define (compile exp cte)

  (define (literal? x)
    (cond ((symbol? x) #f)
          ((pair? x) (eq? (car x) 'quote))
          (else #t)))
  
  (define (varref? x)
    (symbol? x))
  
  (define (conditional? x)
    (eq? (car x) 'if))
  
  (define (lambda? x)
	(eq? (car x) 'named-lambda))
  
  (define (assignment? x)
    (eq? (car x) 'set!))
  
  (define (sequence? x)
    (eq? (car x) 'begin))
  
  (define (let-macro? x)
	(eq? (car x) 'let-macro))

  (define (primop? x)
    (and (symbol? (car x))
         (eq? (string-ref (symbol->string (car x)) 0) #\%)))
  
  (define rest    cdr)
  (define first   car)
  (define second  cadr)
  (define third   caddr)
  (define fourth  cadddr)
  
  ;; (lexical-address '((x y z) (a b c) (d e f)) 'e)
  ;; => '(2 1)
  (define (lexical-address syms var)
    (let dloop ((syms syms) (d 0))
      (if (null? syms)
          #f
          (let ((i (let iloop ((rib (car syms)) (i 0))
                     (if (null? rib)
                         #f
                         (if (eq? (car rib) var)
                             i
                             (iloop (cdr rib) (+ i 1)))))))
            (if i
                (list d i)
                (dloop (cdr syms) (+ d 1)))))))
  
  (define (compile-exp tail? exp lenv senv k)

	(define (macro-reference? exp)
	  (and (symbol? (first exp)) (assq (first exp) senv)))

;	(set! tail? #f)
;	(display (list tail? exp))
;	(newline)

    (cond ((literal? exp) (gen-lit exp k))
  
          ;; ======================================================================
          ((varref? exp)
           (let ((addr (lexical-address lenv exp)))
             (if addr
                 (gen-varref addr k)
				 (compile-dynref tail? exp lenv senv k))))
  
          ;; ======================================================================
          ((assignment? exp)
           (let ((addr (lexical-address lenv (second exp)))
                 (value (third exp)))
             (if addr
                 (compile-exp #f value lenv senv
                  (gen-assign addr k))
                 ;(compile-exp #f value lenv senv
                 ; (gen-dynset addr k))
				 (compile-dynset tail? exp lenv senv k))))
  
          ;; ======================================================================
          ((conditional? exp)
           (let ((test-exp (second exp))
                 (then-exp (third exp))
                 (else-exp (fourth exp)))
             (compile-exp
              #f test-exp lenv senv
              (gen-test
               (compile-exp tail? then-exp lenv senv '())
               (compile-exp tail? else-exp lenv senv '())
               k))))
                         
          ;; ======================================================================
          ((sequence? exp)
           (if (null? (cdr exp))
               (gen-undefined k)
               ;; there's at least one expression
               (let loop ((exprs (cdr exp)))
                 (if (null? (cdr exprs))
                     ;; last expression
                     (compile-exp tail? (car exprs) lenv senv k)
                     ;; non-tail expression
                     (compile-exp #f (car exprs) lenv senv
                      (loop (cdr exprs)))))))
  
          ;; ======================================================================

		  ;; how to compile let-macro.
		  ;; we must compile the body of each transformer *now*, and stuff the
		  ;; closure into <senv>.  This requires an <eval>, doesn't it?

		  ((let-macro? exp)
		   (for-each
			(lambda (m)
			  (let ((name (car m))
					(mdef (cons 'lambda (cdr (cadr m)))))
				(set! senv (cons (cons name (eval mdef (interaction-environment))) senv))))
			(reverse (cadr exp)))
		   (compile-exp tail? `(begin ,@(cddr exp)) lenv senv k))
  
          ;; ======================================================================
          ((lambda? exp)
           (let ((name (second exp))
				 (formals (third exp))
                 (body (if (null? (cddddr exp))
                           (cadddr exp)
                           (cons 'begin (cdddr exp)))))
             (gen-closure
              (compile-exp #t body (cons formals lenv) senv restore-code)
			  name
              k)))
             
          ;; ======================================================================
          ((primop? exp)
           (let ((primop (car exp))
                 (rands (cdr exp)))
             (if (null? rands)
                 (if tail?
                     (gen-primop primop)
                     (gen-save-val (gen-primop primop) k))
                 (let ((primop-code
                        (compile-rands
                         rands lenv senv
                         (gen-primop primop))))
                   (if tail?
                       primop-code
                       (gen-save-val primop-code k))))))
  
          ;; ======================================================================

		  ((macro-reference? exp)
		   (let ((rewrite (cdr (assq (car exp) senv))))
			 ;;(%print (apply rewrite (cdr exp)))
			 ;; compile the transformed code
			 (let ((rewritten (transform (apply rewrite (cdr exp)) (lambda (x) x) #f)))
			   (compile-exp tail? rewritten lenv senv k))))

          ;; ======================================================================
          (else ;; application
           (let ((rator (car exp))
                 (rands (cdr exp)))
             (let ((app-code 
                    (if (and (pair? rator)
							 (lambda? rator))
                        (let ((formals (caddr rator))
                              (body (cdddr rator)))
                          ;; closure analysis: don't generate a closure for
                          ;; forms like this ((lambda (x) ...) ...), just
                          ;; extend the environment and execute the body.
                          (compile-rands
                           rands lenv senv
                           (gen-push-env
                            (compile-exp #f
                             (if (null? (cdr body))
                                 (car body)
                                 (cons 'begin body))
                             (cons formals lenv)
                             senv
                             restore-code))))
                        ;; normal application
                        (compile-rands
                         rands lenv senv
                         (compile-exp #f rator lenv senv invoke-code))
                        )))
               (if tail?
                   app-code
                   (gen-save app-code k)))))
          )
    )
  
  ;; ===========================================================================
  ;;                        Operand Evaluation
  ;; ===========================================================================
  
  (define (compile-rands rands lenv senv k)
    (gen-new-tuple
     (length rands)
     (if (null? rands)
         k ;; no operands, empty tuple
         (compile-tuple-rands rands lenv senv k))))
  
  (define (compile-tuple-rands rands lenv senv k)
    (let loop ((i 0)
               (rands rands))
      (compile-exp
       #f (car rands) lenv senv
       (gen-tuple-store i
        (if (null? (cdr rands))
            k
            (loop (+ i 1) (cdr rands)))))))
  
  ;; ===========================================================================
  ;;                         Insn Generators
  ;; ===========================================================================
  
  (define (gen-lit lit k)
    (cons `(literal ,lit) k))
  
  (define (gen-varref var k)
    (cons `(varref ,var) k))
  
  (define (gen-assign var k)
    (cons `(assign ,var) k))
  
  ;; ~~define, ~~dynset, ~~dynref are provided by the runtime
  ;; to support the dynamic global environment
  
  (define (compile-dynset tail? exp lenv senv k)
	(compile-exp tail? `(~~dynset! (quote ,(second exp)) ,(third exp)) lenv senv k))

  (define (compile-dynref tail? exp lenv senv k)
	(if (eq? exp '~~dynref)
		(error "~~dynref missing from compilation environment")
		(compile-exp tail? `(~~dynref (quote ,exp)) lenv senv k)))

  (define (gen-test then-code else-code k)
    (cons `(test ,then-code ,else-code) k))
  
  (define (gen-primop primop)
    `((primop ,primop) (restore)))
  
  (define (gen-save code k)
    (cons `(save ,code) k))
  
  (define (gen-save-val code k)
    (cons `(save-val ,code) k))
  
  (define (gen-closure code name k)
    (cons `(close ,code ,name) k))
                                   
  (define (gen-new-tuple length k)
    (cons `(new-tuple ,length) k))
  
  (define (gen-tuple-store index k)
    (cons `(tuple-store ,index) k))
  
  (define (gen-undefined k)
    (cons '(undefined #f) k))
  
  (define (gen-push-env k)
    (cons '(push-env) k))
  
  (define restore-code '((restore)))
  (define invoke-code '((invoke)))
  
  ;; don't think this is necessary any longer
  ;;  (let ((expanded (transform exp (lambda (x) x) #f)))
  ;;	(compile-exp #t expanded cte '() restore-code))

  ;; body of compile
  (compile-exp #t exp cte '() restore-code)
  
  )
