;;; -*- Mode: Scheme; tab-width: 4 -*-
;;; the 'eof object' should be a unique object.  why not
;;; just a string?
;;; but who can return the eof object?	not the low-level
;;; routines.

;; buffered i/o

(define eof-object "<the end-of-file object>")

(define (eof-object? x)
  (%eq? x eof-object))

(define (buffered-file fd buffer-size)

  ;; [-----<pos>XXXXXXXXX<end>------]
  
  (let ((buffer (make-string buffer-size))
		(pos 0)
		(end 0))

	(define (read-buffer)
	  (let ((n (%read fd buffer buffer-size)))
		(cond ((%eq? n 0) eof-object)
			  ((> n 0)
			   (set! end n)
			   (set! pos 0))
			  (else
			   (error "%read failed")))))

	(define (write-buffer)
	  (let loop ()
		(if (< pos end)
			(let ((n (%write fd (substring buffer pos end) (%- end pos))))
			  (if (> n 0)
				  (begin (set! pos (+ pos n)) (loop))
				  (error "%write failed")))))
	  (set! pos 0)
	  (set! end 0)
	  )

	(define (read-char)
	  (cond ((< pos end)
			 (let ((ch (%string-ref buffer pos)))
			   (set! pos (%+ pos 1))
			   ch))
			(else
			 (if (%eq? (read-buffer) eof-object)
				 eof-object
				 (read-char)))))

	(define (write-char ch)
	  (cond ((< end buffer-size)
			 (%string-set! buffer end ch)
			 (set! end (%+ end 1)))
			(else
			 (write-buffer)
			 (write-char ch))))

	(define (write-string s)
	  (let ((sl (string-length s)))
		(cond ((%gt? (%+ sl end) buffer-size)
			   ;; just flush then write the string
			   (write-buffer)
			   (%write fd s sl))
			  (else
			   (%buffer-copy s 0 sl buffer end)
			   (set! end (%+ end sl))))))

	(define (close)
	  (flush)
	  (%close fd))

	(define (flush)
	  (write-buffer))

	(dispatcher read-char write-char write-string write-buffer flush close)

	))

(define (open-file filename create? write? mode)
  (let ((zt-filename
		 (string-append filename
						(char->string (integer->char 0)))))
	(let ((fd (%open zt-filename create? write? mode)))
	  (if (< fd 0)
		  (error "%open failed")
		  (buffered-file fd 8192)))))

(define (open-input-file filename)
  (open-file filename #f #f 0))

(define (open-output-file filename)
  (open-file filename #t #t #o644))

(define (close-input-port port)
  ((port 'close)))

(define (close-output-port port)
  ((port 'close)))

(define (read-char port)
  ((port 'read-char)))

(define (write-char ch port)
  ((port 'write-char) ch))

(define (flush port)
  ((port 'flush)))

(define stdin #f)
(define stdout #f)

(define (initialize-stdio)
  ;; unix-specific
  (set! stdin (buffered-file 1 8192))
  (set! stdout (buffered-file 2 8192))
;  (set! print (make-printer stdout))
  )
