;; -*- Mode: Scheme; tab-width: 4 -*-

(define (make-reader port)
  (let ((port-read-char (port 'read-char))
		(char #f)
		(whitespace (map ascii->char '(9 10 12 13 32)))
		(end-of-line (map ascii->char '(10 13)))
		(dec-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
		(oct-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7))
		(hex-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\a #\b #\c #\d #\e #\f))
		)

	(define (peek)
	  (if (eq? char #f)
		  (set! char (port-read-char)))
	  char)

	(define (next)
	  (let ((result char))
		(set! char (port-read-char))
		(if (eq? result #f)
			(next)
			result)))

	(define (skip-whitespace)
	  (let loop ((ch (peek)))
		(cond ((memq ch whitespace) (next) (loop (peek)))
			  ;; comments are treated like whitespace, and must be
			  ;; 'invisible' to the main loop.
			  ((eq? ch #\;) (skip-comment) (loop (peek)))
			  (else ch))))

	(define (skip-comment)
	  (let loop ()
		(if (not (memq (peek) end-of-line))
			(begin (next) (loop)))))

	(define (punctuation? ch)
	  (or (eq? ch #\( )
		  (eq? ch #\) )
		  (memq ch whitespace)))

	(define (unexpected-eof)
	  (error "unexpected end of file!"))

	(define (read-character)
	  (let ((first-char (next)))
		(if (punctuation? (peek))
			first-char
			(let ((name 
				   (let loop ((r (list first-char))
							  (ch (peek)))
					 (if (punctuation? ch)
						 (string->symbol (list->string (reverse r)))
						 (loop (cons (next) r) (peek))))))
			  ;; yet another evil bootstrapping issue. <grin>
			  (cond ((equal? name 'newline)
					 #\newline)
					((equal? name 'space)
					 #\space)
					(else (error "unknown special character-name" name)))))))

	;; simple version, no support for (say) #\newline
	;;(next))

	(define (read-symbol)
	  (let loop ((r (list))
				 (ch (peek)))
		(if (punctuation? ch)
			(string->symbol (list->string (reverse r)))
			(loop (cons (next) r) (peek)))))

	(define (read-string)
	  (let loop ((r (list))
				 (ch (peek)))
		(cond ((eq? ch #\")
			   (next)
			   (list->string (reverse r)))
			  ((eq? ch #\\)
			   (next)
			   (loop (cons (next) r) (peek)))
			  (else
			   (loop (cons (next) r) (peek)))
			  )))

	(define (read-list)
	  (let loop ((r (list)))
		(skip-whitespace)
		(cond ((eq? (peek) #\) )
			   (next)
			   (reverse r))
			  ((eq? (peek) #\. )
			   (next)
			   (skip-whitespace)
			   (let ((tail (read)))
				 (skip-whitespace)
				 (cond ((eq? (peek) #\))
						(next)
						(append (reverse r) tail))
					   (else
						(error "ill-formed list")))))
			  (else (loop (cons (read) r))))))

	(define digit->integer
	  (let ((zero (char->integer #\0)))
		(lambda (ch)
		  (- (char->integer ch) zero))))

	(define (read-number)
	  (let loop ((n (digit->integer (next))))
		(if (memq (peek) dec-digits)
			(loop (+ (* 10 n) (digit->integer (next))))
			(if (punctuation? (peek))
				n
				(error "ill-formed number")))))

	(define (read-hex-number)
	  (let loop ((n 0)
				 (ch (peek)))
		(let ((hd (memq (peek) hex-digits)))
		  (if hd
			  (loop (+ (* 16 n) (- 16 (length hd))) (next))
			  (if (punctuation? (peek))
				  n
				  (error "ill-formed number"))))))

	(define (read-octal-number)
	  (let loop ((n (digit->integer (next))))
		(if (memq (peek) oct-digits)
			(loop (+ (* 8 n) (digit->integer (next))))
			(if (punctuation? (peek))
				n
				(error "ill-formed number")))))

	(define (read)
	  (skip-whitespace)
	  (let ((ch (peek)))
		(cond
		 ((eof-object? ch) ch)
		 ((eq? ch #\( ) (next) (read-list))
		 ((eq? ch #\" ) (next) (read-string))
		 ((eq? ch #\' ) (next) (list 'quote (read)))
		 ((eq? ch #\# )
		  (next)
		  (let ((ch (peek)))
			(cond
			 ((eof-object? ch) (unexpected-eof))
			 ((eq? ch #\\ ) (next) (read-character))
			 ((eq? ch #\t ) (next) #t)
			 ((eq? ch #\f ) (next) #f)
			 ((eq? ch #\( ) (next) (list->vector (read-list)))
			 ((eq? ch #\x ) (next) (read-hex-number))
			 ((eq? ch #\o ) (next) (read-octal-number))
			 (else (error "unsupported read macro" ch)))))
		 ((eq? ch #\- ) (next)
		  (if (memq (peek) dec-digits)
			  (- 0 (read-number))
			  '-))
		 ((eq? ch #\` ) (next) (list 'quasiquote (read)))
		 ((eq? ch #\, ) (next)
		  (cond ((eq? (peek) #\@)
				 (next)
				 (list 'unquote-splicing (read)))
				(else
				 (list 'unquote (read)))))
		 ((eq? ch #\) ) (next) (error "unexpected close-paren"))
		 ((memq ch dec-digits) (read-number))
		 (else (read-symbol)))))

	(define (read-all)
	  (let loop ((all '())
				 (e (read)))
		(if (not (eof-object? e))
			(loop (cons e all) (read))
			(reverse all))))

	(dispatcher read peek next read-all)
	))
