(let ()

;; -*- Mode: Scheme; tab-width: 4 -*-

;; ==================================================
;; comparison
;; ==================================================	 

(define (eq? a b)
  (%eq? a b))

;; I think the main distinction is in comparing numbers,
;; since we only got fixnums, we ignore the difference
;; between eq? and eqv?

(define (eqv? a b)
  (%eq? a b))

(define (equal? a b)
  (cond ((%eq? a b) #t) ;; integers, booleans, symbols, characters
		;; numeric types will break this type-check cheat.
		((%eq? (%get-typecode a) (%get-typecode b))
		 (cond ((string? a)
				(string=? a b))
			   ((pair? a)
				(and (equal? (%car a) (%car b))
					 (equal? (%cdr a) (%cdr b))))
			   ((vector? a)
				(let ((alen (vector-length a))
					  (blen (vector-length b)))
				  (cond ((equal? alen blen)
						 (let loop ((n alen))
						   (cond ((< n 0) #t)
								 ((equal? (vector-ref a (%- n 1))
										  (vector-ref b (%- n 1)))
								  (loop (%- n 1)))
								 (else #f))))
						(else #f))))
			   (else #f)))
		(else #f)))

(define (undefined? x)
  (%eq? (%undefined) x))

;; ==================================================
;; list/pair
;; ==================================================	 

(define (cons a b) (%cons a b))
(define (car x) (%car x))
(define (cdr x) (%cdr x))
(define (caar x) (%car (%car x)))
(define (cadr x) (%car (%cdr x)))
(define (cdar x) (%cdr (%car x)))
(define (cddr x) (%cdr (%cdr x)))
(define (caaar x) (%car (%car (%car x))))
(define (caadr x) (%car (%car (%cdr x))))
(define (cadar x) (%car (%cdr (%car x))))
(define (caddr x) (%car (%cdr (%cdr x))))
(define (cdaar x) (%cdr (%car (%car x))))
(define (cdadr x) (%cdr (%car (%cdr x))))
(define (cddar x) (%cdr (%cdr (%car x))))
(define (cdddr x) (%cdr (%cdr (%cdr x))))
(define (caaaar x) (%car (%car (%car (%car x)))))
(define (caaadr x) (%car (%car (%car (%cdr x)))))
(define (caadar x) (%car (%car (%cdr (%car x)))))
(define (caaddr x) (%car (%car (%cdr (%cdr x)))))
(define (cadaar x) (%car (%cdr (%car (%car x)))))
(define (cadadr x) (%car (%cdr (%car (%cdr x)))))
(define (caddar x) (%car (%cdr (%cdr (%car x)))))
(define (cadddr x) (%car (%cdr (%cdr (%cdr x)))))
(define (cdaaar x) (%cdr (%car (%car (%car x)))))
(define (cdaadr x) (%cdr (%car (%car (%cdr x)))))
(define (cdadar x) (%cdr (%car (%cdr (%car x)))))
(define (cdaddr x) (%cdr (%car (%cdr (%cdr x)))))
(define (cddaar x) (%cdr (%cdr (%car (%car x)))))
(define (cddadr x) (%cdr (%cdr (%car (%cdr x)))))
(define (cdddar x) (%cdr (%cdr (%cdr (%car x)))))
(define (cddddr x) (%cdr (%cdr (%cdr (%cdr x)))))
(define (set-car! l v) (%tuple-set! l 0 v))
(define (set-cdr! l v) (%tuple-set! l 1 v))
(define (null? x) (%eq? x '()))
		 
(define (pair? x)
  (%eq? (%get-typecode x) #x20))

;; incomplete, doesn't check for circular lists

(define (list? l)
  (or (null? l)
	  (and (pair? l) (list? (%cdr l)))))

(define (tuple->list t)
  (let ((len (%tuple-length t)))
	(let loop ((result '())
			   (index (%- len 1)))
	  (if (%eq? index 0)
		  (%cons (%tuple-ref t index) result)
		  (loop (%cons (%tuple-ref t index) result)
				(%- index 1))))))

(define (list->tuple l tag)
  (let ((r (%new-tuple tag (length l))))
	(let loop ((i 0)
			   (l l))
	  (if (%eq? l '())
		  r
		  (begin
			(%tuple-set! r i (%car l))
			(loop (%+ i 1) (%cdr l)))))))

(define (list)
  (%cdr (tuple->list (%get-environment))))

(define (list-ref l i)
  (let loop ((l l) (i i))
	(if (%eq? i 0)
		(%car l)
		(loop (%cdr l) (%- i 1)))))

(define (length l)
  (let loop ((l l)
			 (n 0))
	(if (pair? l)
		(loop (%cdr l) (%+ n 1))
		n)))

(define (append a b)
  (if (null? a)
	  b
	  (let loop ((l a))
		(if (null? (%cdr l))
			(%cons (%car l) b)
			(%cons (%car l) (loop (%cdr l)))))))

(define (reverse l)
  (let loop ((l l)
			 (r '()))
	(if (pair? l)
		(loop (%cdr l) (%cons (%car l) r))
		r)))

(define (member-pred pred)
  (lambda (x l)
	(let loop ((l l))
	  (if (pair? l)
		  (if (pred x (%car l))
			  l
			  (loop (%cdr l)))
		  #f))))
			
(define member #f)
(define memv #f)

(define (memq x l) (%memq x l))
(define (assq x l) (%assq x l))

(define (assoc-pred pred)
  (lambda (x l)
	(let loop ((l l))
	  (if (pair? l)
		  (if (pred x (%car (%car l)))
			  (%car l)
			  (loop (%cdr l)))
		  #f))))

;; set below
(define assoc-string #f)

;; ==================================================
;; symbols
;; ==================================================	 

;; There is only one symbol table.	 Its job is to maintain
;; the singular property of symbols: that two symbols based
;; on the same string will compare equal (with eq?).
;;
;; Note that the symbol-table and (dynamic) environment are
;; not the same thing.	 Environments use symbols as their
;; keys.

(define symbol-table #f)

(define (symbol? a)
  (%eq? (%get-typecode a) #x24))

(define (string->uninterned-symbol s)
  (let ((t (%new-tuple #x24 1)))
	(%tuple-set! t 0 s)
	t))

(define (string-downcase s)
  (list->string (map char-downcase (string->list s))))

(define (string->symbol str)
  (let ((str (string-downcase str)))
	(let ((interned? (assoc-string str symbol-table)))
	  (if interned?
		  ;; if such a symbol already exists, return it.
		  (%cdr interned?)
		  ;; otherwise, intern a new symbol into the table and return it.
		  (let ((symbol (string->uninterned-symbol str)))
			(set! symbol-table (%cons (%cons str symbol) symbol-table))
			symbol)))))

(define (symbol->string s)
  (%tuple-ref s 0))

(define (intern-symbol s)
  (set! symbol-table (%cons (%cons (symbol->string s) s) symbol-table)))

(define gensym
  (let ((n 0))
	(lambda ()
	  (set! n (+ n 1))
	  (string->symbol (string-append "g" (number->string n))))))

;; ==================================================
;; dynamic environments
;; ==================================================

(define (error . args)
  (%print "Error: " )
  (%print args))

(define verbose-interaction #t)

(define print
  (lambda args

	(define (w x)
	  (if (string? x)
		  ((stdout 'write-string) x)
		  (write* x stdout)))

	(let loop ((l args))
	  (cond ((null? l) #f)
			((null? (cdr l))
			 (w (car l))
			 (loop (cdr l)))
			(else
			 (w (car l))
			 ((stdout 'write-string) " ")
			 (loop (cdr l)))))
	((stdout 'write-char) (integer->char 10))
	((stdout 'flush))))

(define (make-printer port)

  (lambda args

	(define (w x)
	  (if (string? x)
		  ((port 'write-string) x)
		  (write* x port)))

	(let loop ((l args))
	  (cond ((null? l) #f)
			((null? (cdr l))
			 (w (car l))
			 (loop (cdr l)))
			(else
			 (w (car l))
			 ((port 'write-string) " ")
			 (loop (cdr l)))))
	((port 'write-char) (integer->char 10))
	((port 'flush))))

;(define print #f)

;; defined below
(define the-environment #f)
(define ~~dynset! #f)
(define ~~dynref #f)

;; ==================================================
;; strings
;; ==================================================	 

(define (string-length s)
  (%tuple-ref s 0))

(define (string? x)
  (%eq? (%get-typecode x) #x14))

(define (string=? a b)
  (%eq? (%string-compare a b) 0))

(define (string<? a b)
  (%eq? (%string-compare a b) (%- 0 1)))

(define (string>? a b)
  (%eq? (%string-compare a b) 1))
			
(define (string->list s)
  (let loop ((n (string-length s))
			 (r '()))
	(if (%eq? n 0)
		r
		(loop (%- n 1) (%cons (%string-ref s (%- n 1)) r)))))

(define (string-fill! s ch)
  (let loop ((n (string-length s)))
	(if (< n 0)
		#f
		(begin
		  (%string-set! s n ch)
		  (loop (%- n 1))))))

(define (make-string n)
  (if (check-room n)
	  (let ((s (%new-tuple #x14 (%+ (how-many n 4) 1))))
		(%tuple-set! s 0 n)
		s
		)
	  (error "out of heap for string of size" n)))

(define (string-append1 a b)
  (let ((alen (string-length a))
		(blen (string-length b)))
	(let ((r (make-string (%+ alen blen))))
	  (let aloop ((i 0))
		(if (%lt? i alen)
			(begin (%string-set! r i (string-ref a i))
				   (aloop (%+ i 1)))))
	  (let bloop ((i 0))
		(if (%lt? i blen)
			(begin (%string-set! r (%+ i alen) (string-ref b i))
				   (bloop (%+ i 1)))))
	  r)))

;; this probably wants to be a builtin
(define (string-append first . rest)
  (let loop ((r first)
			 (l rest))
	(if (null? l)
		r
		(loop (string-append1 r (%car l)) (%cdr l)))))

(define (substring string start end)
  (let ((n (%- end start)))
	(let ((r (make-string n)))
	  (let loop ((i 0))
		(if (%eq? i n)
			r
			(loop (begin
					(%string-set! r i (%string-ref string (%+ i start)))
					(%+ i 1))))))))

(define (list->string l)
  (let ((n (length l)))
	(let ((s (make-string n)))
	  (let loop ((i 0)
				 (l l))
		(cond ((null? l) s)
			  (else
			   (%string-set! s i (%car l))
			   (loop (%+ i 1) (%cdr l))))))))

(define (char->string ch)
  (let ((s (make-string 1)))
	(%string-set! s 0 ch)
	s))

(define (string . chars)
  (list->string chars))

(define (string-ref s i)
  (%string-ref s i))

(define (string-set! s i c)
  (%string-set! s i c))

(define (zero-terminate s)
  (string-append s (char->string (ascii->char 0))))

;;(define (buffer-copy src src-start n dst dst-start)
;;  (let loop ((i 0))
;;	(when (%lt? i n)
;;	 (string-set! dst (%+ dst-start i) (%string-ref src (%+ src-start i)))
;;	 (loop (%+ i 1)))))

(define (buffer-copy src src-start n dst dst-start)
  (%buffer-copy src src-start n dst dst-start))

;; ==================================================
;; characters
;; ==================================================	 

(define (integer->char n)
  (%integer->char n))

(define (char->integer ch)
  (%char->integer ch))

(define (ascii->char n)
  (%integer->char n))

(define (char->ascii ch)
  (%char->integer ch))

(define (char? x)
  (%eq? (%get-typecode x) #x02))

(define char-downcase
  (let ((A (%char->integer #\A))
		(Z (%char->integer #\Z)))
	(lambda (ch)
	  (let ((code (%char->integer ch)))
		(if (and (%ge? code A) (%le? code Z))
			(ascii->char (%+ code 32))
			ch)))))

;; ==================================================
;; vectors
;; ==================================================	 

(define (vector? x)
  (%eq? (%get-typecode x) #x18))

(define (vector-ref v i)
  (%tuple-ref v i))

(define (vector-set! v i d)
  (%tuple-set! v i d))

(define (vector-length v)
  (%tuple-length v))

(define (vector-map! p v)
  (let loop ((n (vector-length v)))
	(if (%eq? n 0)
		#f
		(begin
		  (vector-set! v (%- n 1) (p (vector-ref v (%- n 1))))
		  (loop (%- n 1))))))

(define (make-vector n)
  (let ((args (%get-environment))
		(nargs (%- (%tuple-length (%get-environment)) 1))
		(fill #f))
	(if (> nargs 1)
		(set! fill (%tuple-ref args 2)))
	(let ((result (%new-tuple #x18 n)))
	  (vector-map! (lambda (x) fill) result)
	  result)))

(define (vector->list v)
  (let loop ((n (%- (vector-length v) 1))
			 (r '()))
	(if (< n 0)
		r
		(loop (%- n 1) (%cons (vector-ref v n) r)))))

(define (list->vector l)
  (list->tuple l #x18))

(define (vector . l)
  (list->vector l))

;; ==================================================
;; boolean
;; ==================================================	 

(define (boolean? x)
  (%eq? (%get-typecode x) #x06))

(define (not x)
  (%eq? x #f))

;; ==================================================
;; numbers
;; ==================================================	 

(define (+ a b) (%+ a b))
(define (- a b) (%- a b))
(define (* a b) (%* a b))
(define (/ a b) (%/ a b))

(define (= a b) (%eq? a b))
(define (> a b) (%gt? a b))
(define (< a b) (%lt? a b))
(define (>= a b) (%ge? a b))
(define (<= a b) (%le? a b))

(define (zero? x) (= x 0))
(define (negative? x) (< x 0))

(define (quotient a b)
  (%/ a b))

(define (remainder a b)
  (%% a b))
		 
(define (even? x)
  (zero? (remainder x 2)))

(define (odd? x)
  (not (even? x)))

(define (abs x)
  (if (negative? x) (%- 0 x) x))

(define (modulo a b)
  (let ((r (remainder a b)))
	(if (= r 0)
		0
		(if (< a 0)
			(if (< b 0) r (%+ r b))
			(if (< b 0) (%+ r b) r)))))

(define (expt n e)
  (let loop ((i e)
			 (a n))
	(if (eq? e 0)
		a
		(loop (%- i 1) (* a n)))))

(define (number? x)
  (%eq? (%get-typecode x) #x00))

(define (integer? x)
  (%eq? (%get-typecode x) #x00))

(define (number->string n . base)
  (let ((table "0123456789abcdefghijklmnopqrstuvwxyz"))
	(if (null? base)
		(set! base 10)
		(set! base (car base)))
	(if (zero? n)
		"0"
		(let loop ((x (abs n))
				   (r '()))
		  (if (zero? x)
			  (list->string
			   (if (negative? n) (%cons #\- r) r))
			  (loop (quotient x base)
					(%cons (%string-ref table (remainder x base)) r)))))))

;; ==================================================
;; control flow
;; ==================================================

(define (procedure? p)
  (%eq? (%get-typecode p) #xc0))

(define (apply proc args)
  ;; leave room for env link for <extend-env>
  (%apply1 proc (list->tuple (%cons 0 args) #x1c)))

(define (map1 p l)
  (let loop ((l l) (r '()))
	(if (pair? l)
		(loop (%cdr l) (%cons (p (%car l)) r))
		(reverse r)
		)))

(define (map* p l)
  (let loop ((l l) (r '()))
	(if (null? (%car l))
		(reverse r)
		(loop (map1 cdr l)
			  (%cons (apply p (map1 car l)) r))
		)))

;; varargs original
(define (map p . l)
  (if (null? (cdr l))
	  (map1 p (car l))
	  (map* p l)))

 ;; really for-each1
(define (for-each p l)
  (let loop ((l l))
	(if (null? l)
		0
		(begin (p (%car l))
			   (loop (%cdr l))))))

(define (call/cc p)
  (let ((k (%getcc)))
	(p (lambda (result)
		 (%putcc k result)))))

(define (compile-to-closure exp env inline-primitives)
  (let ((tc (transform exp (lambda (x) x) inline-primitives)))
	(let ((cs (compile-cvm tc (car env))))
	  (let ((co (%build-tuple #x28 (list->vector (%cdr cs)) (%car cs))))
		(%build-tuple #x0c co 0 (cdr env))))))

;; taken from repl.scm
(define (eval exp . env)
  (if (null? env)
	  (set! env (interaction-environment))
	  (set! env (car env)))
  (let ((closure (compile-to-closure exp env #f)))
	(closure)))

;; an environment for eval really consists of the
;; the symbol/depth bindings and the actual environment
;; tuple that goes with it.
(define (interaction-environment)
  (cons *cts* *cte*))

;; ==================================================
;; input/output
;; ==================================================	 

;; not really
(define (display x)
  (%print x))

;; to be replaced by write.scm::write
(define (write x)
  (%print x))

(define (newline . port)
  (if (null? port)
	  (set! port stdin)
	  (set! port (car port)))
  ((port 'write-char) (integer->char 10)))

(define (read . port)
  (if (null? port)
	  (set! port stdin)
	  (set! port (car port)))
  (((make-reader port) 'read)))

(define (load filename)
  (let ((port (open-input-file filename)))
	(let ((reader (make-reader port)))
	  (if verbose-interaction (print "; read..."))
	  (let ((forms ((reader 'read-all))))
		((port 'close))
		(if verbose-interaction (print "; compile..."))
		(for-each eval forms);; top-level defines will by dynamic
		))))

(define (read-forms-from-file filename)
  (let ((port (open-input-file filename)))
	(let ((reader (make-reader port)))
	  (let ((forms ((reader 'read-all))))
		((port 'close))
		forms))))

;; ==================================================
;; misc
;; ==================================================	 

(define (check-room nbytes)
  (if (< (%room) nbytes)
	  (>= (%gc) nbytes)
	  #t))
					  
(define (how-many x n)
  (%/ (%- (%+ x n) 1) n))

(define (tuple? x)
  (%eq? (%get-typecode x) #x1c))

;; last result (this should dynamic?)
(define _ #f)
;; compile-time-environment
(define *cte* #f)
;; compile-time symbol-table
(define *cts* #f)

(define (make-environment)
  (let ((table (list)))
	
	(define (new-cell sym)
	  (let ((cell (%cons sym (%undefined))))
		(set! table (%cons cell table))
		cell))

	(define (dynref sym)
	  (let ((probe (%assq sym table)))
		(if probe
			(%cdr probe)
			(error "undefined dynamic symbol" sym))))

	(define (dynset! sym val)
	  (let ((probe (%assq sym table)))
		(let ((cell (if probe probe (new-cell sym))))
		  ;; (set-cdr! cell val)
		  (%tuple-set! cell 1 val) ;; until we get macros
		  )))

	(%cons dynref dynset!)))

;; this is a generic object-dispatcher, it builds a closure
;; that will map symbols to 'methods':
;; (dispatcher one two)
;; =>
;; (let ((table `((one . ,one)
;;                (two . ,two))))
;;   (lambda (msg)
;;     (let ((probe (%assq msg table)))
;;       (if probe
;;           (%cdr probe)
;;           (error "no such method")))))

(define-macro (dispatcher . methods)
  (let ((pairs (list 'quasiquote (map (lambda (name) (cons name (list 'unquote name))) methods)))
		(table (gensym))
		(msg (gensym))
		(probe (gensym)))
	`(let ((,table ,pairs))
	   (lambda (,msg)
		 (let ((,probe (%assq ,msg ,table)))
		   (if ,probe
			   (%cdr ,probe)
			   (error "no such method")))))))

(define (initialize-runtime)

  (%write-string "initializing runtime...") (%write-char (integer->char 10))

  ;; setup

  ;; compile-time-environment
  (set! *cte* (%tuple-ref (%get-environment) 0))
  (set! symbol-table (list))
  (set! assoc-string (assoc-pred string=?))
  (set! member (member-pred equal?))
  (set! memv (member-pred eqv?))

  (set! the-environment (make-environment))
  ;; these are referenced by compiled code
  (set! ~~dynref  (%car the-environment))
  (set! ~~dynset! (%cdr the-environment))

  ;; KLUDGE
  ;; Bootstrap puts a vector of symbol objects where we can get at it.
  (let ((top-env (%tuple-ref (%tuple-ref (%get-environment) 0) 0)))
	(let ((symbols (%tuple-ref top-env 0)))
	  (vector-map! intern-symbol symbols)
	  ;; clear it
	  (%tuple-set! top-env 0 '())
	  ))

  ;; KLUDGE
  ;; Bootstrap puts the compile-time symbols at the end of the code
  ;; object's constant vector.
  
  (let ((cv (%tuple-ref (%tuple-ref eq? 0) 0)))
	(let ((vl (vector-length cv)))
	  (set! *cts* (list '() (vector-ref cv (%- vl 1))))
	  ;; clear it
	  (%tuple-set! cv (%- vl 1) #f)
	  ))
  
  (%write-string "done.") (%write-char (integer->char 10))
  )

;; -*- Mode: Scheme; tab-width: 4 -*-

(define (make-reader port)
  (let ((port-read-char (port 'read-char))
		(char #f)
		(whitespace (map ascii->char '(9 10 12 13 32)))
		(end-of-line (map ascii->char '(10 13)))
		(dec-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
		(oct-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7))
		(hex-digits '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\a #\b #\c #\d #\e #\f))
		)

	(define (peek)
	  (if (eq? char #f)
		  (set! char (port-read-char)))
	  char)

	(define (next)
	  (let ((result char))
		(set! char (port-read-char))
		(if (eq? result #f)
			(next)
			result)))

	(define (skip-whitespace)
	  (let loop ((ch (peek)))
		(cond ((memq ch whitespace) (next) (loop (peek)))
			  ;; comments are treated like whitespace, and must be
			  ;; 'invisible' to the main loop.
			  ((eq? ch #\;) (skip-comment) (loop (peek)))
			  (else ch))))

	(define (skip-comment)
	  (let loop ()
		(if (not (memq (peek) end-of-line))
			(begin (next) (loop)))))

	(define (punctuation? ch)
	  (or (eq? ch #\( )
		  (eq? ch #\) )
		  (memq ch whitespace)))

	(define (unexpected-eof)
	  (error "unexpected end of file!"))

	(define (read-character)
	  (let ((first-char (next)))
		(if (punctuation? (peek))
			first-char
			(let ((name 
				   (let loop ((r (list first-char))
							  (ch (peek)))
					 (if (punctuation? ch)
						 (string->symbol (list->string (reverse r)))
						 (loop (cons (next) r) (peek))))))
			  ;; yet another evil bootstrapping issue. <grin>
			  (cond ((equal? name 'newline)
					 #\newline)
					((equal? name 'space)
					 #\space)
					(else (error "unknown special character-name" name)))))))

	;; simple version, no support for (say) #\newline
	;;(next))

	(define (read-symbol)
	  (let loop ((r (list))
				 (ch (peek)))
		(if (punctuation? ch)
			(string->symbol (list->string (reverse r)))
			(loop (cons (next) r) (peek)))))

	(define (read-string)
	  (let loop ((r (list))
				 (ch (peek)))
		(cond ((eq? ch #\")
			   (next)
			   (list->string (reverse r)))
			  ((eq? ch #\\)
			   (next)
			   (loop (cons (next) r) (peek)))
			  (else
			   (loop (cons (next) r) (peek)))
			  )))

	(define (read-list)
	  (let loop ((r (list)))
		(skip-whitespace)
		(cond ((eq? (peek) #\) )
			   (next)
			   (reverse r))
			  ((eq? (peek) #\. )
			   (next)
			   (skip-whitespace)
			   (let ((tail (read)))
				 (skip-whitespace)
				 (cond ((eq? (peek) #\))
						(next)
						(append (reverse r) tail))
					   (else
						(error "ill-formed list")))))
			  (else (loop (cons (read) r))))))

	(define digit->integer
	  (let ((zero (char->integer #\0)))
		(lambda (ch)
		  (- (char->integer ch) zero))))

	(define (read-number)
	  (let loop ((n (digit->integer (next))))
		(if (memq (peek) dec-digits)
			(loop (+ (* 10 n) (digit->integer (next))))
			(if (punctuation? (peek))
				n
				(error "ill-formed number")))))

	(define (read-hex-number)
	  (let loop ((n 0)
				 (ch (peek)))
		(let ((hd (memq (peek) hex-digits)))
		  (if hd
			  (loop (+ (* 16 n) (- 16 (length hd))) (next))
			  (if (punctuation? (peek))
				  n
				  (error "ill-formed number"))))))

	(define (read-octal-number)
	  (let loop ((n (digit->integer (next))))
		(if (memq (peek) oct-digits)
			(loop (+ (* 8 n) (digit->integer (next))))
			(if (punctuation? (peek))
				n
				(error "ill-formed number")))))

	(define (read)
	  (skip-whitespace)
	  (let ((ch (peek)))
		(cond
		 ((eof-object? ch) ch)
		 ((eq? ch #\( ) (next) (read-list))
		 ((eq? ch #\" ) (next) (read-string))
		 ((eq? ch #\' ) (next) (list 'quote (read)))
		 ((eq? ch #\# )
		  (next)
		  (let ((ch (peek)))
			(cond
			 ((eof-object? ch) (unexpected-eof))
			 ((eq? ch #\\ ) (next) (read-character))
			 ((eq? ch #\t ) (next) #t)
			 ((eq? ch #\f ) (next) #f)
			 ((eq? ch #\( ) (next) (list->vector (read-list)))
			 ((eq? ch #\x ) (next) (read-hex-number))
			 ((eq? ch #\o ) (next) (read-octal-number))
			 (else (error "unsupported read macro" ch)))))
		 ((eq? ch #\- ) (next)
		  (if (memq (peek) dec-digits)
			  (- 0 (read-number))
			  '-))
		 ((eq? ch #\` ) (next) (list 'quasiquote (read)))
		 ((eq? ch #\, ) (next)
		  (cond ((eq? (peek) #\@)
				 (next)
				 (list 'unquote-splicing (read)))
				(else
				 (list 'unquote (read)))))
		 ((eq? ch #\) ) (next) (error "unexpected close-paren"))
		 ((memq ch dec-digits) (read-number))
		 (else (read-symbol)))))

	(define (read-all)
	  (let loop ((all '())
				 (e (read)))
		(if (not (eof-object? e))
			(loop (cons e all) (read))
			(reverse all))))

	(dispatcher read peek next read-all)
	))
;; -*- Mode: Scheme; tab-width: 4 -*-

(define gensym
  (let ((n 0))
	(lambda ()
	  (set! n (+ n 1))
	  (string->symbol (string-append "g" (number->string n))))))


(define transform

  (let ((inline-substitutions
		 '((cons			%cons)
		   (car				%car)
		   (cdr				%cdr)
		   (eq?				%eq?)
		   (+				%+)
		   (-				%-)
		   (*				%*)
		   (quotient		%/)
		   (=				%eq?)
		   (integer->char	%integer->char)
		   (char->integer	%char->integer)
		   (assq			%assq)
		   (memq			%memq)
		   ))
		(transform-alist '())
		)
	;;
	;; One eyeball on MzScheme's transformers.
	;;
	;; Thanks to Dorai Sitaram and Matthew Flatt for assistance.
	;;
	;; let, let*, letrec, and, or, cond, begin, define, quasiquote,
	;; unquote, unquote-splicing, do, case
	
	(define (raise-syntax-error . l)
	  (error l))

	(define gensym
	  (let ((n 0))
		(lambda ()
		  (set! n (+ n 1))
		  (string->symbol (string-append "g" (number->string n))))))

	;;
	;; The Tricky Bits:
	;; 1) definitions inside bodies.
	;; 2) variable-length argument lists (two forms)
	;;
	;; These nasty warts must be handled within the source tree walker.
	;;

	;; (lambda x <body>)
	;; =>
	;; (let ((x (%cdr (tuple->list (%get_environment))))) <body>)
	;; => ((lambda (x) <body>) (%cdr (tuple->list (%get_environment))))
	;;
	;; (lambda (x y z . l) <body>)
	;; =>
	;; (let ((g23 (%cdr (tuple->list (%get_environment)))))
	;;    (let ((x (%car g23))
	;;          (y (%car (%cdr g23)))
	;;          (z (%car (%cdr (%cdr g23))))
	;;          (l (%cdr (%cdr (%cdr g23)))))
	;;       <body>))
	;; [the %cdr/%car chains are inline primitives, thus the code is
	;;  actually small compared to calling something like 'cadddr']
	;; TODO: we should probably recast this in terms of %tuple-ref.

	(define frob-definition
	  (lambda (e)
		(cond ((pair? (car e))
			   `(,(caar e) (named-lambda ,(caar e) (,@(cdr (car e))) ,@(cdr e))))
			  ((symbol? (car e)) e)
			  (else
			   (raise-syntax-error "bad syntax (define)" e)))))

	(define (frob-varargs e)

	  (define (frob-dotted formals body)
		(let ((sym (gensym)))
		  `(lambda ()
			 (let ((,sym (%cdr (tuple->list (%get-environment)))))
			   (let ,(let loop ((l formals)
								(r '())
								(a sym))
					   (cond ((pair? l)
							  (loop (cdr l)
									(cons `(,(car l) (%car ,a)) r)
									`(%cdr ,a)))
							 (else (reverse (cons `(,l ,a) r)))))
				 ,@body)))))
	
	  (let ((formals (car e))
			(body (cdr e)))
		(cond ((symbol? formals)
			   `(lambda () ((lambda (,formals) ,@body) (%cdr (tuple->list (%get-environment))))))
			  (else (frob-dotted formals body)))))

	(define (collect-matching-forms seq sym)
	  ;; collect forms from <seq> starting with <sym>
	  (let loop ((body seq)
				 (non-matches '())
				 (matches '()))
		(cond ((null? body)
			   (cons non-matches matches))
			  ((and (pair? (car body)) (eq? (caar body) sym))
			   (loop (cdr body) non-matches (cons (car body) matches)))
			  (else
			   (loop (cdr body) (cons (car body) non-matches) matches))
			  )))

	(define (frob-macro m)
	  (let ((formals (cadr m))
			(body (cddr m)))
		(let ((name (car formals))
			  (formals (cdr formals)))
		  `(,name (compile-time-lambda ,formals ,@body)))))

	(define (frob-body e)
	  ;; first, pull out macro definitions
	  (let ((scan (collect-matching-forms e 'define-macro)))
		(let ((forms (car scan))
			  (macrodefs (cdr scan)))
		  ;; then pull out definitions
		  (let ((scan (collect-matching-forms forms 'define)))
			(let ((forms (car scan))
				  (definitions (cdr scan)))
			  (let ((forms
					 (if (null? definitions)
						 forms
						 `((letrec ,(map frob-definition (map cdr definitions)) ,@forms)))))
				(if (not (null? macrodefs))
					`((let-macro
					   ;; (define-macro (<name> <arg0> <arg1> ... ) <body>)
					   ;; => (letmacro ((name (lambda (<arg0> <arg1) <body>))) ...)
					   ,(map frob-macro macrodefs)
					   (begin ,@forms)))
					forms)))))))
			  
	(define (initialize-transforms)

	  (define (undefined)
		(%undefined))

	  (define (define-transform name proc)
		(set! transform-alist (cons (cons name proc) transform-alist)))

	  (define-transform 'and
		(lambda (e)
		  (cond ((null? e) #t)
				((null? (cdr e)) (car e))
				(else `(if ,(car e) (and ,@(cdr e)))))))

	  (define-transform 'or
		(lambda (e)
		  (cond ((null? e) #f)
				((null? (cdr e)) (car e))
				(else
				 (let ((temp (gensym)))
				   `(let ((,temp ,(car e)))
					  (if ,temp
						  ,temp
						  (or ,@(cdr e))))
				   )))))

	  ;; (when <test> <exp1> <exp2> ...)
	  ;; => (if <test>
	  ;;       (begin <exp1> <exp2> ...))

	  (define-transform 'when
		(lambda (e)
		  `(if ,(car e) (begin ,@(cdr e)))))

	  ;; (let <variable0> ((<variable1> <init1>) ...)
	  ;;      <body>)
	  ;; ; =>
	  ;; ((letrec ((<variable0> (lambda (<variable1> ...) <body>)))
	  ;;    <variable0>)
	  ;;  <init1> ...)
	  ;;
	  ;; (let ((<name1> <init1>)
	  ;;       (<name2> <init2>) ...)
	  ;;    <body>)
	  ;;  =>
	  ;; ((lambda (<name1> <name2> ...) <body>) <init1> <init2>)
  
	  (define-transform 'let
		(lambda (e)
		  (if (symbol? (car e))
			  ;; named let
			  (let ((loopvar (car e))
					(vars (map car (cadr e)))
					(inits (map cadr (cadr e)))
					(body (cddr e)))
				`((letrec ((,loopvar (lambda ,vars ,@body)))
					,loopvar)
				  ,@inits))
			  ;; normal let
			  (let ((decls (car e))
					(body (cdr e)))
				(let ((names (map car decls))
					  (inits (map cadr decls)))
				  `((lambda ,names ,@body) ,@inits))))))


	  ;; (let* ((<name1> <init1>)
	  ;;        (<name2> <init2>)
	  ;;        ...)
	  ;;   <body>)
	  ;; =>
	  ;; (let ((<name1> <init1>))
	  ;;   (let* ((<name2> <init2>))
	  ;;      <body>))

	  (define-transform 'let*
		(lambda (e)
		  (let ((decls (car e)))
			(if (null? (cdr decls))
				;; one binding
				(cons 'let e)
				;; two+ bindings
				`(let (,(car decls))
				   (let* (,@(cdr decls)) ,@(cdr e)))))))

	  ;; (letrec ((<name1> <init1)
	  ;;          (<name2> <init2) ... )
	  ;;   <body>)
	  ;; =>
	  ;; (let ((<name1> #f)
	  ;;       (<name2> #f) ... )
	  ;;   (set! <name1> <init1>)
	  ;;   (set! <name2> <init2>)
	  ;;   <body>)
  
	  (define-transform 'letrec
		(lambda (e)
		  (let ((decls (car e))
				(body (cdr e)))
			(let ((empty-decls (map (lambda (d) `(,(car d) #f)) decls))
				  (assigns (map (lambda (d) `(set! ,(car d) ,(cadr d))) decls)))
			  `(let (,@empty-decls)
				 ,@assigns
				 ,@body)))))


	  ;; top-level define
	  (define-transform 'define
		(lambda (e)
		  (let ((fe (frob-definition e)))
			`(~~dynset! (quote ,(car fe)) ,@(cdr fe)))))

	  ;;(define-transform 'define
	  ;;  (lambda (e)
	  ;;	(cons 'set! (frob-definition e))))
  
	  (define-transform 'cond
		(lambda (form)
		  (let loop ((tests form))
			(if (null? tests)
				'(%undefined)
				(if (not (pair? tests))
					(raise-syntax-error "bad syntax (body must contain a list of pairs)" tests)
					(let ((line (car tests))
						  (rest (cdr tests)))
					  (if (not (pair? line))
						  (raise-syntax-error "bad syntax (clause is not a test-value pair)" line)
						  (let* ((test (car line))
								 (value (cdr line))
								 (else? (eq? test 'else)))
							(if (and else? (pair? rest))
								(raise-syntax-error "bad syntax (`else' clause must be last)" line))
							(if (and (pair? value)
									 (eq? '=> (car value)))
								(if (and (pair? (cdr value))
										 (null? (cddr value)))
									(let ((test (if else?
													#t 
													test))
										  (gen (gensym)))
									  `(let ((,gen ,test))
										 (if ,gen
											 (,(cadr value) ,gen)
											 (loop rest))))
									(raise-syntax-error "bad syntax (bad clause form with =>)" line))
								(if else?
									(cons 'begin value)
									(if (null? value)
										(let ((gen (gensym)))
										  `(let ((,gen ,test))
											 (if ,gen ,gen ,(loop rest))))
										(list 'if test (cons 'begin value) (loop rest)))))))))))))

	  ;; taken from JAR's simple-macro package

	  (define-transform 'quasiquote

        (lambda (x)

          (define (expand-quasiquote x level)
            (descend-quasiquote x level finalize-quasiquote))

          (define (finalize-quasiquote mode arg)
            (cond ((eq? mode 'quote) (list 'quote arg))
                  ((eq? mode 'unquote) arg)
                  ((eq? mode 'unquote-splicing)
                   (syntax-error ",@ in invalid context" arg))
                  (else `(,mode ,@arg))))

          (define (descend-quasiquote x level return)
            (cond ((vector? x)
                   (descend-quasiquote-vector x level return))
                  ((not (pair? x))
                   (return 'quote x))
                  ((interesting-to-quasiquote? x 'quasiquote)
                   (descend-quasiquote-pair x (+ level 1) return))
                  ((interesting-to-quasiquote? x 'unquote)
                   (cond ((= level 0)
                          (return 'unquote (cadr x)))
                         (else
                          (descend-quasiquote-pair x (- level 1) return))))
                  ((interesting-to-quasiquote? x 'unquote-splicing)
                   (cond ((= level 0)
                          (return 'unquote-splicing (cadr x)))
                         (else
                          (descend-quasiquote-pair x (- level 1) return))))
                  (else
                   (descend-quasiquote-pair x level return))))

          (define (descend-quasiquote-pair x level return)
            (descend-quasiquote (car x) level
              (lambda (car-mode car-arg)
                (descend-quasiquote (cdr x) level
                  (lambda (cdr-mode cdr-arg)
                    (cond ((and (eq? car-mode 'quote) (eq? cdr-mode 'quote))
                           (return 'quote x))
                          ((eq? car-mode 'unquote-splicing)
                           ;; (,@mumble ...)
                           (cond ((and (eq? cdr-mode 'quote) (null? cdr-arg))
                                  (return 'unquote
                                          car-arg))
                                 (else
                                  (return 'append
                                          (list car-arg (finalize-quasiquote
                                                           cdr-mode cdr-arg))))))
                          (else
                           (return 'cons
                                   (list (finalize-quasiquote car-mode car-arg)
                                         (finalize-quasiquote cdr-mode cdr-arg))))))))))

          (define (descend-quasiquote-vector x level return)
            (descend-quasiquote (vector->list x) level
              (lambda (mode arg)
                (case mode
                  ((quote) (return 'quote x))
                  (else (return 'list->vector
                                (list (finalize-quasiquote mode arg))))))))

          (define (interesting-to-quasiquote? x marker)
            (and (pair? x) (eq? (car x) marker)))

          (expand-quasiquote (car x) 0)))

	  (define-transform 'case
		(lambda (form)
		  (let ((v (car form))
				(tests-in (cdr form))
				(gen (gensym)))
			`(let ((,gen ,v))
			   (cond
				,@(let loop ((tests tests-in))
					(cond
					 ((null? tests) '())
					 ((pair? tests)
					  (let ((test-value (car tests))
							(rest (cdr tests)))
						(cons
						 (if (or (not (pair? test-value))
								 (not (pair? (cdr test-value))))
							 (raise-syntax-error "bad syntax (clause is not a test-value pair)" test-value)
							 (let ((test (car test-value))
								   (value (cdr test-value)))
							   (if (and (pair? value) (list? value))
								   #t
								   (raise-syntax-error "bad syntax (improper clause body)" value))
							   (cons
								(cond
								 ((eq? test 'else)
								  (if (null? rest)
									  'else
									  (raise-syntax-error "bad syntax (`else' clause must be last)" test-value)))
								 (else
								  (cons 
								   'or
								   (let loop ((test test))
									 (cond
									  ((null? test) '())
									  ((pair? test)
									   (let ((t (list 'quote (car test))))
										 (cons `(eqv? ,gen ,t)
											   (loop (cdr test)))))
									  (else (loop (cons test '()))))))))
								`((begin ,@value)))))
						 (loop rest))))
					 (else (raise-syntax-error "bad syntax (body must contain a list of pairs)" tests)))))))))

	  (define-transform 'do
		(lambda (form)
		  (let ((bindings (car form))
				(test-and-result (cadr form))
				(body (cddr form)))
			(let ((code (cons 'do form)))
			  (if (list? bindings)
				  (undefined)
				  (raise-syntax-error 'do "bad syntax (bad binding sequence)" code))
			  (for-each (lambda (binding)
						  (if (and (list? binding)
								   (let ((l (length binding)))
									 (or (= l 2) (= l 3))))
							  (undefined)
							  (raise-syntax-error 'do "bad syntax (bad binding)" code)))
						bindings)
			  (if (list? test-and-result) (undefined)
				  (raise-syntax-error 'do "bad syntax (bad test and result)" code))
			  (if (list? body) (undefined)
				  (raise-syntax-error 'do "bad syntax (improper body)" code))
			  (let ((vars (map (lambda (binding)
								 (let ((var (car binding)))
								   (if (symbol? var)
									   var
									   (raise-syntax-error 'do "bad syntax (variable must be an identifier)" code))))
							   bindings))
					(inits (map cadr bindings))
					(steps (map (lambda (binding)
								  (cond
								   ((pair? (cddr binding))
									(caddr binding))
								   (else (car binding))))
								bindings))
					(test (if (pair? test-and-result)
							  (car test-and-result)
							  (raise-syntax-error 'do "bad syntax (test/result part)" code)))
					(results (if (null? (cdr test-and-result))
								 (list '(undefined))
								 (cdr test-and-result)))
					(loop (gensym)))
				`(let ,loop ,(map (lambda (var init)
									(list var init))
								  vars inits)
					  (if ,test
						  (begin
							,@results)
						  (begin
							,@body
							(,loop ,@steps)))))))))

	  )

	(initialize-transforms)

	(lambda (e p inline-primitives)

	  (let ((substitutions (if inline-primitives
							   inline-substitutions
							   '())))

		(define (varargs? l)
		  (or (symbol? l)
			  (let loop ((l l))
				(cond ((null? l) #f)
					  ((pair? l) (loop (cdr l)))
					  (else #t)))))

		(define (transform* e)
		  (p
		   (if (pair? e)
			   (let ((front (car e)))
				 (case front
				   ;; conditional
				   ((if)
					(if (= (length e) 3)
						`(if ,(transform* (cadr e)) ,(transform* (caddr e)) #f);; #spam
						`(if ,(transform* (cadr e)) ,(transform* (caddr e)) ,(transform* (cadddr e)))))
				   ;; assignment
				   ((set!) `(set! ,(cadr e) ,(transform* (caddr e))))
				   ;; procedure
				   ((named-lambda)
					(if (varargs? (caddr e))
						(transform* (frob-varargs (cddr e)))
						`(named-lambda ,(cadr e) ,(caddr e) ,@(map transform* (frob-body (cdddr e))))))
				   ((lambda)
					(if (varargs? (cadr e))
						(transform* (frob-varargs (cdr e)))
						`(named-lambda #f ,(cadr e) ,@(map transform* (frob-body (cddr e))))))
				   ;; compile-time-lambda
				   ;; this is a lambda that will be eval'd at compile-time.
				   ;; leave it alone.
				   ((compile-time-lambda)
					e)
				   ;; sequence
				   ((begin)
					(if (= (length e) 2)
						(transform* (cadr e))
						(cons 'begin (map transform* (cdr e)))))
				   ;; literal data
				   ((quote)
					e)
				   (else
					(if (symbol? front)
						;; macro
						(let ((macro-probe (assq front transform-alist)))
						  (if macro-probe
							  (transform* ((cdr macro-probe) (cdr e)))
							  (let ((subst-probe (assq front substitutions)))
								(if subst-probe 
									(cons (cadr subst-probe) (map transform* (cdr e)))
									(map transform* e)))))
						;; application
						(map transform* e)))))
			   e)))
		(transform* e))))
	)
;; -*- Mode: Scheme; tab-width: 4 -*-

(define (compile exp cte)

  (define (literal? x)
    (cond ((symbol? x) #f)
          ((pair? x) (eq? (car x) 'quote))
          (else #t)))
  
  (define (varref? x)
    (symbol? x))
  
  (define (conditional? x)
    (eq? (car x) 'if))
  
  (define (lambda? x)
	(eq? (car x) 'named-lambda))
  
  (define (assignment? x)
    (eq? (car x) 'set!))
  
  (define (sequence? x)
    (eq? (car x) 'begin))
  
  (define (let-macro? x)
	(eq? (car x) 'let-macro))

  (define (primop? x)
    (and (symbol? (car x))
         (eq? (string-ref (symbol->string (car x)) 0) #\%)))
  
  (define rest    cdr)
  (define first   car)
  (define second  cadr)
  (define third   caddr)
  (define fourth  cadddr)
  
  ;; (lexical-address '((x y z) (a b c) (d e f)) 'e)
  ;; => '(2 1)
  (define (lexical-address syms var)
    (let dloop ((syms syms) (d 0))
      (if (null? syms)
          #f
          (let ((i (let iloop ((rib (car syms)) (i 0))
                     (if (null? rib)
                         #f
                         (if (eq? (car rib) var)
                             i
                             (iloop (cdr rib) (+ i 1)))))))
            (if i
                (list d i)
                (dloop (cdr syms) (+ d 1)))))))
  
  (define (compile-exp tail? exp lenv senv k)

	(define (macro-reference? exp)
	  (and (symbol? (first exp)) (assq (first exp) senv)))

;	(set! tail? #f)
;	(display (list tail? exp))
;	(newline)

    (cond ((literal? exp) (gen-lit exp k))
  
          ;; ======================================================================
          ((varref? exp)
           (let ((addr (lexical-address lenv exp)))
             (if addr
                 (gen-varref addr k)
				 (compile-dynref tail? exp lenv senv k))))
  
          ;; ======================================================================
          ((assignment? exp)
           (let ((addr (lexical-address lenv (second exp)))
                 (value (third exp)))
             (if addr
                 (compile-exp #f value lenv senv
                  (gen-assign addr k))
                 ;(compile-exp #f value lenv senv
                 ; (gen-dynset addr k))
				 (compile-dynset tail? exp lenv senv k))))
  
          ;; ======================================================================
          ((conditional? exp)
           (let ((test-exp (second exp))
                 (then-exp (third exp))
                 (else-exp (fourth exp)))
             (compile-exp
              #f test-exp lenv senv
              (gen-test
               (compile-exp tail? then-exp lenv senv '())
               (compile-exp tail? else-exp lenv senv '())
               k))))
                         
          ;; ======================================================================
          ((sequence? exp)
           (if (null? (cdr exp))
               (gen-undefined k)
               ;; there's at least one expression
               (let loop ((exprs (cdr exp)))
                 (if (null? (cdr exprs))
                     ;; last expression
                     (compile-exp tail? (car exprs) lenv senv k)
                     ;; non-tail expression
                     (compile-exp #f (car exprs) lenv senv
                      (loop (cdr exprs)))))))
  
          ;; ======================================================================

		  ;; how to compile let-macro.
		  ;; we must compile the body of each transformer *now*, and stuff the
		  ;; closure into <senv>.  This requires an <eval>, doesn't it?

		  ((let-macro? exp)
		   (for-each
			(lambda (m)
			  (let ((name (car m))
					(mdef (cons 'lambda (cdr (cadr m)))))
				(set! senv (cons (cons name (eval mdef (interaction-environment))) senv))))
			(reverse (cadr exp)))
		   (compile-exp tail? `(begin ,@(cddr exp)) lenv senv k))
  
          ;; ======================================================================
          ((lambda? exp)
           (let ((name (second exp))
				 (formals (third exp))
                 (body (if (null? (cddddr exp))
                           (cadddr exp)
                           (cons 'begin (cdddr exp)))))
             (gen-closure
              (compile-exp #t body (cons formals lenv) senv restore-code)
			  name
              k)))
             
          ;; ======================================================================
          ((primop? exp)
           (let ((primop (car exp))
                 (rands (cdr exp)))
             (if (null? rands)
                 (if tail?
                     (gen-primop primop)
                     (gen-save-val (gen-primop primop) k))
                 (let ((primop-code
                        (compile-rands
                         rands lenv senv
                         (gen-primop primop))))
                   (if tail?
                       primop-code
                       (gen-save-val primop-code k))))))
  
          ;; ======================================================================

		  ((macro-reference? exp)
		   (let ((rewrite (cdr (assq (car exp) senv))))
			 ;;(%print (apply rewrite (cdr exp)))
			 ;; compile the transformed code
			 (let ((rewritten (transform (apply rewrite (cdr exp)) (lambda (x) x) #f)))
			   (compile-exp tail? rewritten lenv senv k))))

          ;; ======================================================================
          (else ;; application
           (let ((rator (car exp))
                 (rands (cdr exp)))
             (let ((app-code 
                    (if (and (pair? rator)
							 (lambda? rator))
                        (let ((formals (caddr rator))
                              (body (cdddr rator)))
                          ;; closure analysis: don't generate a closure for
                          ;; forms like this ((lambda (x) ...) ...), just
                          ;; extend the environment and execute the body.
                          (compile-rands
                           rands lenv senv
                           (gen-push-env
                            (compile-exp #f
                             (if (null? (cdr body))
                                 (car body)
                                 (cons 'begin body))
                             (cons formals lenv)
                             senv
                             restore-code))))
                        ;; normal application
                        (compile-rands
                         rands lenv senv
                         (compile-exp #f rator lenv senv invoke-code))
                        )))
               (if tail?
                   app-code
                   (gen-save app-code k)))))
          )
    )
  
  ;; ===========================================================================
  ;;                        Operand Evaluation
  ;; ===========================================================================
  
  (define (compile-rands rands lenv senv k)
    (gen-new-tuple
     (length rands)
     (if (null? rands)
         k ;; no operands, empty tuple
         (compile-tuple-rands rands lenv senv k))))
  
  (define (compile-tuple-rands rands lenv senv k)
    (let loop ((i 0)
               (rands rands))
      (compile-exp
       #f (car rands) lenv senv
       (gen-tuple-store i
        (if (null? (cdr rands))
            k
            (loop (+ i 1) (cdr rands)))))))
  
  ;; ===========================================================================
  ;;                         Insn Generators
  ;; ===========================================================================
  
  (define (gen-lit lit k)
    (cons `(literal ,lit) k))
  
  (define (gen-varref var k)
    (cons `(varref ,var) k))
  
  (define (gen-assign var k)
    (cons `(assign ,var) k))
  
  ;; ~~define, ~~dynset, ~~dynref are provided by the runtime
  ;; to support the dynamic global environment
  
  (define (compile-dynset tail? exp lenv senv k)
	(compile-exp tail? `(~~dynset! (quote ,(second exp)) ,(third exp)) lenv senv k))

  (define (compile-dynref tail? exp lenv senv k)
	(if (eq? exp '~~dynref)
		(error "~~dynref missing from compilation environment")
		(compile-exp tail? `(~~dynref (quote ,exp)) lenv senv k)))

  (define (gen-test then-code else-code k)
    (cons `(test ,then-code ,else-code) k))
  
  (define (gen-primop primop)
    `((primop ,primop) (restore)))
  
  (define (gen-save code k)
    (cons `(save ,code) k))
  
  (define (gen-save-val code k)
    (cons `(save-val ,code) k))
  
  (define (gen-closure code name k)
    (cons `(close ,code ,name) k))
                                   
  (define (gen-new-tuple length k)
    (cons `(new-tuple ,length) k))
  
  (define (gen-tuple-store index k)
    (cons `(tuple-store ,index) k))
  
  (define (gen-undefined k)
    (cons '(undefined #f) k))
  
  (define (gen-push-env k)
    (cons '(push-env) k))
  
  (define restore-code '((restore)))
  (define invoke-code '((invoke)))
  
  ;; don't think this is necessary any longer
  ;;  (let ((expanded (transform exp (lambda (x) x) #f)))
  ;;	(compile-exp #t expanded cte '() restore-code))

  ;; body of compile
  (compile-exp #t exp cte '() restore-code)
  
  )
;; -*- Mode: Scheme; tab-width: 4 -*-

(define (make-bytecoder)
  (let ((opcodes
		 '((literal				. #\L)
		   (constant			. #\C)
		   (varref				. #\V)
		   (assign				. #\!)
		   (save				. #\S)
		   (restore				. #\R)
		   (save-val			. #\q)
		   (test				. #\?)
		   (primop				. #\%)
		   (invoke				. #\I)
		   (jump				. #\J)
		   ;;(close				. #\P)
		   (close				. #\newline)
		   (halt				. #\H)
		   (new-tuple			. #\T)
		   (tuple-store			. #\>)
		   (push-env			. #\Z)
		   (pushcc				. #\K)
		   (loadcc				. #\k)
		   (builtin				. #\B)
		   ;; (store0-9			. #\0-#\9)
		   ;; (lexref[0-3][0-3]	. #\a-p)
		   ))
		(primops
		 '((%+					. #\+)
		   (%-					. #\-)
		   (%*					. #\*)
		   (%/					. #\/)
		   (%%					. #\%)
		   (%eq?				. #\=)
		   (%gt?				. #\>)
		   (%lt?				. #\<)
		   (%ge?				. #\})
		   (%le?				. #\{)
		   (%land				. #\&)
		   (%lor				. #\|)
		   (%xor				. #\x)
		   ;(%print				. #\$)
		   ;; minimal additional primops:
		   ;; cons car cdr
		   ;; integer->char char->integer
		   ;; string-ref string-set!
		   ;; new-tuple tuple-length tuple-ref tuple-set!
		   ;; gc room
		   (%cons				. #\a)
		   (%car				. #\b)
		   (%cdr				. #\c)
		   (%integer->char		. #\d)
		   (%char->integer		. #\e)
		   (%apply1				. #\f)
		   ;;(%string-set!		. #\g)
		   (%new-tuple			. #\h)
		   (%tuple-length		. #\i)
		   (%tuple-ref			. #\j)
		   (%tuple-set!			. #\k)
		   (%callcc				. #\l)
		   (%build-tuple		. #\m)
		   (%undefined			. #\n)
		   (%gc					. #\o)
		   (%dump-image			. #\p)
		   (%get-typecode		. #\q)
		   (%get-environment	. #\r)
		   (%room				. #\s)
		   (%insns				. #\t)
		   (%getcc				. #\u)
		   (%putcc				. #\v)
		   (%left-shift			. #\w)
		   (%right-shift		. #\y)
		   ;; i386 builtins:
		   ;; print
		   ;; dump-regs
		   ;; get-typecode
		   ;; string-compare
		   ;; make-closure
		   ;; getc putc
		   ;; open read write close
		   ;; dump
		   ;; get-platform open-library close-library get-symbol call-foreign-function
		   ))
		(builtins
		 '(%print
		   %read-char
		   %write-char
		   %string-compare
		   %string-ref
		   %string-set!
		   %buffer-copy
		   %write-string
		   %flush
		   %assq
		   %memq
		   %open
		   %read
		   %write
		   %close
		   %socket
		   %connect
		   %bind
		   %listen
		   %accept
		   %clock
		   ))
		 
		(integer-table '#(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\A #\B #\C #\D #\E #\F))
		(lexref-table '#(#\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m #\n #\o #\p))
		(new-tuple-table '#(#\. #\; #\" #\,))

		)

	(define rest    cdr)
	(define first   car)
	(define second  cadr)
	(define third   caddr)
	(define fourth  cadddr)

	(lambda (code)
	  (let ((constants '())
			(n-constants 0))

		(define (new-constant c)

		  (define (index c)
			(let loop ((i 1)
					   (l constants))
			  (if (null? l)
				  #f
				  (if (equal? c (car l))
					  (- n-constants i)
					  (loop (+ i 1) (cdr l))))))

		  (let ((probe (index c)))
			(if probe
				probe
				(begin
				  (set! constants (cons c constants))
				  (set! n-constants (+ n-constants 1))
				  (- n-constants 1)))))

		(define (emit-code code)

		  ;; We have two different ways to encode integers in the VM stream,
		  ;; this way is more compact but slightly more expensive to unpack.
		  ;; It also makes the bytecode more readable.

		  (define (encode-integer n)
		    (let ((neg (negative? n)))
		      (let loop ((n (if neg (- n) n))
		                 (r (if neg '(#\-) '(#\+))))
		        (if (zero? n)
		            r
		            (loop (quotient n #x10)
		                  (cons (vector-ref integer-table (modulo n #x10)) r))))))
		  
;		  (define (encode-integer n)
;			(let loop ((n n)
;					   (i 0)
;					   (r '()))
;			  (if (eq? i 4)
;				  (map ascii->char (reverse r))
;				  (loop (quotient n 256) (+ i 1) (cons (modulo n 256) r)))))

		  (define (encode-literal x)
			;; should this check be done elsewhere?
			(if (and (pair? x) (eq? (car x) 'quote))
				;; ok, this is wrong.  it needs to be (quote <literal>)
				(let ((datum (cadr x)))
				  ;; objects that eval to themselves
				  (if (or (number? datum)
						  (string? datum)
						  (char? datum)
						  (boolean? datum))
					  (encode-literal datum)
					  (list->string (cons #\C (encode-integer (new-constant datum))))))
				(let ((num
					   (cond ((integer? x) (+ 1 (* 2 x)))
							 ((boolean? x) (if x #x106 #x06))
							 ((char? x) (+ #x02 (* #x100 (char->ascii x))))
							 ((null? x) #x0a)
							 (else #f))))
				  (if num
					  (list->string (cons #\L (encode-integer num)))
					  (list->string (cons #\C (encode-integer (new-constant x)))))
				  )))

		  (define (lookup opcode)
			(cdr (assq opcode opcodes)))
		  
		  (define (lookup-primop name)
			(assq name primops))

		  (define (lookup-builtin name)
			(let loop ((i 0)
					   (l builtins))
			  (if (null? l)
				  #f
				  (if (eq? name (car l))
					  i
					  (loop (+ i 1) (cdr l))))))

		  (define (emit-insn insn)
			(let ((opcode (car insn)))
			  (let ((opchar (lookup opcode)))
				(case opcode
				  ((literal)
				   (encode-literal (second insn)))
				  ((primop)
				   ;; we split the notion of primops and builtins here
				   ;; at the compiler level, we don't see a difference;
				   ;; but at the VM level, we do.
				   (let ((pprobe (lookup-primop (second insn))))
					 (if pprobe
						 (list->string (list opchar (cdr pprobe)))
						 (let ((bprobe (lookup-builtin (second insn))))
						   (if bprobe
							   (list->string (cons #\B (encode-integer bprobe)))
							   (error "unknown primop/builtin")))
						 )))
				  ((new-tuple)
				   (let ((size (second insn)))
					 (if (< size 4)
						 (list->string (list (vector-ref new-tuple-table size)))
						 (list->string (cons opchar (encode-integer (second insn)))))))
				  
				  ;; originally, OP_TUPLE_STORE included an index operand
				  ;; since the tuple is always populated in order,
				  ;; we can leave the operand out.  The VM uses the (temporarily unused)
				  ;; 'next' pointer as an index into the 'next' tuple slot to be assigned.
				  ;; this shrinks the bytecode by about 20%, and speeds up the VM.
				  ;; optimized OP_STORE
				  ;;((tuple-store)
				  ;;(string opchar))
				  ;; Ok, that optimization is cool, but it violates the condition
				  ;; that continuations are immutable, and thus breaks call/cc
				  ;; Here's a better version
				  ((tuple-store)
				   (let ((index (second insn)))
					 (if (< index 10)
						 (number->string index)
						 (list->string (cons opchar (encode-integer (second insn)))))))
				  ((test)
				   (let ((then-code (emit-code (second insn)))
						 (else-code (emit-code (third insn))))
					 (let ((jump-code
							(list->string
							 (cons (lookup 'jump)
								   (encode-integer (string-length else-code))))))
					   (string-append
						(list->string
						 (cons opchar
							   (encode-integer (+ (string-length jump-code) (string-length then-code)))))
						then-code
						jump-code
						else-code
						))))
				  ((close)
				   (let ((proc-code (emit-code (second insn)))
						 (name (third insn)))
					 (string-append
					  (list->string (cons opchar (encode-integer (string-length proc-code))))
					  proc-code)))
				  ((save)
				   (let ((proc-code (emit-code (second insn))))
					 (string-append
					  (list->string (cons opchar (encode-integer (string-length proc-code))))
					  proc-code)))
				  ((save-val)
				   (let ((proc-code (emit-code (second insn))))
					 (string-append (string opchar) proc-code)))
				  ((varref assign)
				   (let ((depth (first (second insn)))
						 (index (second (second insn))))
					 (if (and (eq? opcode 'varref) (< depth 4) (< index 4))
						 (list->string (list (vector-ref lexref-table (+ (* 4 depth) index))))
						 (list->string
						  (append
						   (cons opchar (encode-integer (first (second insn))))
						   (encode-integer (second (second insn))))))))
				  (else
				   (string opchar))
				  ))))

		  (apply string-append (map emit-insn code))
		  )
		(cons (emit-code code) (reverse constants))
		))))

(define (compile-cvm exp cte)
  ((make-bytecoder) (compile exp cte)))
;;; -*- Mode: Scheme; tab-width: 4 -*-
;;; the 'eof object' should be a unique object.  why not
;;; just a string?
;;; but who can return the eof object?	not the low-level
;;; routines.

;; buffered i/o

(define eof-object "<the end-of-file object>")

(define (eof-object? x)
  (%eq? x eof-object))

(define (buffered-file fd buffer-size)

  ;; [-----<pos>XXXXXXXXX<end>------]
  
  (let ((buffer (make-string buffer-size))
		(pos 0)
		(end 0))

	(define (read-buffer)
	  (let ((n (%read fd buffer buffer-size)))
		(cond ((%eq? n 0) eof-object)
			  ((> n 0)
			   (set! end n)
			   (set! pos 0))
			  (else
			   (error "%read failed")))))

	(define (write-buffer)
	  (let loop ()
		(if (< pos end)
			(let ((n (%write fd (substring buffer pos end) (%- end pos))))
			  (if (> n 0)
				  (begin (set! pos (+ pos n)) (loop))
				  (error "%write failed")))))
	  (set! pos 0)
	  (set! end 0)
	  )

	(define (read-char)
	  (cond ((< pos end)
			 (let ((ch (%string-ref buffer pos)))
			   (set! pos (%+ pos 1))
			   ch))
			(else
			 (if (%eq? (read-buffer) eof-object)
				 eof-object
				 (read-char)))))

	(define (write-char ch)
	  (cond ((< end buffer-size)
			 (%string-set! buffer end ch)
			 (set! end (%+ end 1)))
			(else
			 (write-buffer)
			 (write-char ch))))

	(define (write-string s)
	  (let ((sl (string-length s)))
		(cond ((%gt? (%+ sl end) buffer-size)
			   ;; just flush then write the string
			   (write-buffer)
			   (%write fd s sl))
			  (else
			   (%buffer-copy s 0 sl buffer end)
			   (set! end (%+ end sl))))))

	(define (close)
	  (flush)
	  (%close fd))

	(define (flush)
	  (write-buffer))

	(dispatcher read-char write-char write-string write-buffer flush close)

	))

(define (open-file filename create? write? mode)
  (let ((zt-filename
		 (string-append filename
						(char->string (integer->char 0)))))
	(let ((fd (%open zt-filename create? write? mode)))
	  (if (< fd 0)
		  (error "%open failed")
		  (buffered-file fd 8192)))))

(define (open-input-file filename)
  (open-file filename #f #f 0))

(define (open-output-file filename)
  (open-file filename #t #t #o644))

(define (close-input-port port)
  ((port 'close)))

(define (close-output-port port)
  ((port 'close)))

(define (read-char port)
  ((port 'read-char)))

(define (write-char ch port)
  ((port 'write-char) ch))

(define (flush port)
  ((port 'flush)))

(define stdin #f)
(define stdout #f)

(define (initialize-stdio)
  ;; unix-specific
  (set! stdin (buffered-file 1 8192))
  (set! stdout (buffered-file 2 8192))
;  (set! print (make-printer stdout))
  )
;; -*- Mode: Scheme; tab-width: 4 -*-

;; not finished.  eventually to replace %print, so we
;; can do output to raw descriptors.

(define (write* obj port)

  (let ((port-write-char (port 'write-char))
	(port-write-string (port 'write-string)))
    
    (define (write-string s)
      (let ((sl (string-length s)))
		(port-write-char #\")
		(let loop ((i 0))
		  (if (%lt? i sl)
			  (let ((ch (%string-ref s i)))
				(if (%eq? ch #\")
					(port-write-char #\\))
				(port-write-char ch)
				(loop (%+ i 1)))))
		(port-write-char #\")))

	(define (write-list l)
	  (port-write-char #\()
	  (let loop ((l l))
		(let ((first (%car l))
			  (rest (%cdr l)))
		  (write-object first)
		  (cond ((null? rest)
				 (port-write-char #\)))
				((pair? rest)
				 (port-write-char #\ )
				 (loop rest))
				(else
				 (port-write-string " . ")
				 (write-object rest)
				 (port-write-char #\)))))))

	(define (write-vector v)
	  (let ((vl (vector-length v)))
		(port-write-char #\#)
		(port-write-char #\()
		(let loop ((i 0))
		  (when (%lt? i vl)
				(write-object (vector-ref v i))
				(if (%lt? i (%- vl 1))
					(port-write-char #\ ))
				(loop (%+ i 1))))
		(port-write-char #\))))

	(define (write-number n)
	  (port-write-string (number->string n)))

	(define (write-object obj)
	  (cond ((char? obj)
			 (port-write-char #\#)
			 (port-write-char #\\)
			 (port-write-char obj))
			((boolean? obj)
			 (port-write-char #\#)
			 (if obj
				 (port-write-char #\t)
				 (port-write-char #\f)))
			((number? obj) (write-number obj))
			((null? obj) (port-write-string "()"))
			((undefined? obj) (port-write-string "#{undefined}"))
			((pair? obj) (write-list obj))
			((vector? obj) (write-vector obj))
			((string? obj) (write-string obj))
			((symbol? obj) (port-write-string (symbol->string obj)))
			(else
			 (port-write-char #\{)
			 (port-write-string
			  (case (%get-typecode obj)
				((#x0c) "closure ")
				((#x10) "save ")
				((#x1c) "tuple ")
				((#x28) "code ")
				(else "unknown-type ")))
			 (port-write-string (number->string (%- obj 0) #x10))
			 (port-write-char #\})
			 )
			))

	(write-object obj)
	))
(initialize-runtime)
(initialize-stdio)
;; if you uncomment the next line be sure to uncomment the line
;; "test-runtime.scm" in the list of runtime sources in s48-bootstrap.scm.
;;(test-runtime)
((stdout 'write-string) "Lunacy.  Copyright 1998-2000 Sam Rushing")
(newline stdout)

(define (repl level in-port out-port)
  (let ((print (make-printer out-port))
	(old-error error))

    (set! error
	  (lambda args
	    (print args)
	    (repl (+ level 1) stdin stdout)))

    (define (prompt)
      (if (> level 0)
	  (write* level out-port))
      ((out-port 'write-string) "> ")
      ((out-port 'flush)))

    (prompt)

    (let loop ((exp (read in-port)))
      (if (not (eof-object? exp))
	  (let ((insns (%insns)))
	    (let ((closure (compile-to-closure exp (interaction-environment) #t)))
	      (let ((cinsns (%- (%insns) insns)))
		(let ((insns (%insns)))
		  (let ((result (closure)))
		    (let ((einsns (%- (%insns) insns)))
		      (if verbose-interaction
			  (print "; compile: " cinsns " execute: " einsns))
		      (write* result out-port)
		      (newline out-port)
		      ((out-port 'flush))
		      (set! _ result)
		      (prompt)
		      (loop (read in-port))))))))))
    (set! error old-error)
    ;; return the last value
    _
    ))

;; interact
(repl 0 stdin stdout)
)

