;; -*- Mode: Scheme; tab-width: 4 -*-

;; ==================================================
;; comparison
;; ==================================================	 

(define (eq? a b)
  (%eq? a b))

;; I think the main distinction is in comparing numbers,
;; since we only got fixnums, we ignore the difference
;; between eq? and eqv?

(define (eqv? a b)
  (%eq? a b))

(define (equal? a b)
  (cond ((%eq? a b) #t) ;; integers, booleans, symbols, characters
		;; numeric types will break this type-check cheat.
		((%eq? (%get-typecode a) (%get-typecode b))
		 (cond ((string? a)
				(string=? a b))
			   ((pair? a)
				(and (equal? (%car a) (%car b))
					 (equal? (%cdr a) (%cdr b))))
			   ((vector? a)
				(let ((alen (vector-length a))
					  (blen (vector-length b)))
				  (cond ((equal? alen blen)
						 (let loop ((n alen))
						   (cond ((< n 0) #t)
								 ((equal? (vector-ref a (%- n 1))
										  (vector-ref b (%- n 1)))
								  (loop (%- n 1)))
								 (else #f))))
						(else #f))))
			   (else #f)))
		(else #f)))

(define (undefined? x)
  (%eq? (%undefined) x))

;; ==================================================
;; list/pair
;; ==================================================	 

(define (cons a b) (%cons a b))
(define (car x) (%car x))
(define (cdr x) (%cdr x))
(define (caar x) (%car (%car x)))
(define (cadr x) (%car (%cdr x)))
(define (cdar x) (%cdr (%car x)))
(define (cddr x) (%cdr (%cdr x)))
(define (caaar x) (%car (%car (%car x))))
(define (caadr x) (%car (%car (%cdr x))))
(define (cadar x) (%car (%cdr (%car x))))
(define (caddr x) (%car (%cdr (%cdr x))))
(define (cdaar x) (%cdr (%car (%car x))))
(define (cdadr x) (%cdr (%car (%cdr x))))
(define (cddar x) (%cdr (%cdr (%car x))))
(define (cdddr x) (%cdr (%cdr (%cdr x))))
(define (caaaar x) (%car (%car (%car (%car x)))))
(define (caaadr x) (%car (%car (%car (%cdr x)))))
(define (caadar x) (%car (%car (%cdr (%car x)))))
(define (caaddr x) (%car (%car (%cdr (%cdr x)))))
(define (cadaar x) (%car (%cdr (%car (%car x)))))
(define (cadadr x) (%car (%cdr (%car (%cdr x)))))
(define (caddar x) (%car (%cdr (%cdr (%car x)))))
(define (cadddr x) (%car (%cdr (%cdr (%cdr x)))))
(define (cdaaar x) (%cdr (%car (%car (%car x)))))
(define (cdaadr x) (%cdr (%car (%car (%cdr x)))))
(define (cdadar x) (%cdr (%car (%cdr (%car x)))))
(define (cdaddr x) (%cdr (%car (%cdr (%cdr x)))))
(define (cddaar x) (%cdr (%cdr (%car (%car x)))))
(define (cddadr x) (%cdr (%cdr (%car (%cdr x)))))
(define (cdddar x) (%cdr (%cdr (%cdr (%car x)))))
(define (cddddr x) (%cdr (%cdr (%cdr (%cdr x)))))
(define (set-car! l v) (%tuple-set! l 0 v))
(define (set-cdr! l v) (%tuple-set! l 1 v))
(define (null? x) (%eq? x '()))
		 
(define (pair? x)
  (%eq? (%get-typecode x) #x20))

;; incomplete, doesn't check for circular lists

(define (list? l)
  (or (null? l)
	  (and (pair? l) (list? (%cdr l)))))

(define (tuple->list t)
  (let ((len (%tuple-length t)))
	(let loop ((result '())
			   (index (%- len 1)))
	  (if (%eq? index 0)
		  (%cons (%tuple-ref t index) result)
		  (loop (%cons (%tuple-ref t index) result)
				(%- index 1))))))

(define (list->tuple l tag)
  (let ((r (%new-tuple tag (length l))))
	(let loop ((i 0)
			   (l l))
	  (if (%eq? l '())
		  r
		  (begin
			(%tuple-set! r i (%car l))
			(loop (%+ i 1) (%cdr l)))))))

(define (list)
  (%cdr (tuple->list (%get-environment))))

(define (list-ref l i)
  (let loop ((l l) (i i))
	(if (%eq? i 0)
		(%car l)
		(loop (%cdr l) (%- i 1)))))

(define (length l)
  (let loop ((l l)
			 (n 0))
	(if (pair? l)
		(loop (%cdr l) (%+ n 1))
		n)))

(define (append a b)
  (if (null? a)
	  b
	  (let loop ((l a))
		(if (null? (%cdr l))
			(%cons (%car l) b)
			(%cons (%car l) (loop (%cdr l)))))))

(define (reverse l)
  (let loop ((l l)
			 (r '()))
	(if (pair? l)
		(loop (%cdr l) (%cons (%car l) r))
		r)))

(define (member-pred pred)
  (lambda (x l)
	(let loop ((l l))
	  (if (pair? l)
		  (if (pred x (%car l))
			  l
			  (loop (%cdr l)))
		  #f))))
			
(define member #f)
(define memv #f)

(define (memq x l) (%memq x l))
(define (assq x l) (%assq x l))

(define (assoc-pred pred)
  (lambda (x l)
	(let loop ((l l))
	  (if (pair? l)
		  (if (pred x (%car (%car l)))
			  (%car l)
			  (loop (%cdr l)))
		  #f))))

;; set below
(define assoc-string #f)

;; ==================================================
;; symbols
;; ==================================================	 

;; There is only one symbol table.	 Its job is to maintain
;; the singular property of symbols: that two symbols based
;; on the same string will compare equal (with eq?).
;;
;; Note that the symbol-table and (dynamic) environment are
;; not the same thing.	 Environments use symbols as their
;; keys.

(define symbol-table #f)

(define (symbol? a)
  (%eq? (%get-typecode a) #x24))

(define (string->uninterned-symbol s)
  (let ((t (%new-tuple #x24 1)))
	(%tuple-set! t 0 s)
	t))

(define (string-downcase s)
  (list->string (map char-downcase (string->list s))))

(define (string->symbol str)
  (let ((str (string-downcase str)))
	(let ((interned? (assoc-string str symbol-table)))
	  (if interned?
		  ;; if such a symbol already exists, return it.
		  (%cdr interned?)
		  ;; otherwise, intern a new symbol into the table and return it.
		  (let ((symbol (string->uninterned-symbol str)))
			(set! symbol-table (%cons (%cons str symbol) symbol-table))
			symbol)))))

(define (symbol->string s)
  (%tuple-ref s 0))

(define (intern-symbol s)
  (set! symbol-table (%cons (%cons (symbol->string s) s) symbol-table)))

(define gensym
  (let ((n 0))
	(lambda ()
	  (set! n (+ n 1))
	  (string->symbol (string-append "g" (number->string n))))))

;; ==================================================
;; dynamic environments
;; ==================================================

(define (error . args)
  (%print "Error: " )
  (%print args))

(define verbose-interaction #t)

(define print
  (lambda args

	(define (w x)
	  (if (string? x)
		  ((stdout 'write-string) x)
		  (write* x stdout)))

	(let loop ((l args))
	  (cond ((null? l) #f)
			((null? (cdr l))
			 (w (car l))
			 (loop (cdr l)))
			(else
			 (w (car l))
			 ((stdout 'write-string) " ")
			 (loop (cdr l)))))
	((stdout 'write-char) (integer->char 10))
	((stdout 'flush))))

(define (make-printer port)

  (lambda args

	(define (w x)
	  (if (string? x)
		  ((port 'write-string) x)
		  (write* x port)))

	(let loop ((l args))
	  (cond ((null? l) #f)
			((null? (cdr l))
			 (w (car l))
			 (loop (cdr l)))
			(else
			 (w (car l))
			 ((port 'write-string) " ")
			 (loop (cdr l)))))
	((port 'write-char) (integer->char 10))
	((port 'flush))))

;(define print #f)

;; defined below
(define the-environment #f)
(define ~~dynset! #f)
(define ~~dynref #f)

;; ==================================================
;; strings
;; ==================================================	 

(define (string-length s)
  (%tuple-ref s 0))

(define (string? x)
  (%eq? (%get-typecode x) #x14))

(define (string=? a b)
  (%eq? (%string-compare a b) 0))

(define (string<? a b)
  (%eq? (%string-compare a b) (%- 0 1)))

(define (string>? a b)
  (%eq? (%string-compare a b) 1))
			
(define (string->list s)
  (let loop ((n (string-length s))
			 (r '()))
	(if (%eq? n 0)
		r
		(loop (%- n 1) (%cons (%string-ref s (%- n 1)) r)))))

(define (string-fill! s ch)
  (let loop ((n (string-length s)))
	(if (< n 0)
		#f
		(begin
		  (%string-set! s n ch)
		  (loop (%- n 1))))))

(define (make-string n)
  (if (check-room n)
	  (let ((s (%new-tuple #x14 (%+ (how-many n 4) 1))))
		(%tuple-set! s 0 n)
		s
		)
	  (error "out of heap for string of size" n)))

(define (string-append1 a b)
  (let ((alen (string-length a))
		(blen (string-length b)))
	(let ((r (make-string (%+ alen blen))))
	  (let aloop ((i 0))
		(if (%lt? i alen)
			(begin (%string-set! r i (string-ref a i))
				   (aloop (%+ i 1)))))
	  (let bloop ((i 0))
		(if (%lt? i blen)
			(begin (%string-set! r (%+ i alen) (string-ref b i))
				   (bloop (%+ i 1)))))
	  r)))

;; this probably wants to be a builtin
(define (string-append first . rest)
  (let loop ((r first)
			 (l rest))
	(if (null? l)
		r
		(loop (string-append1 r (%car l)) (%cdr l)))))

(define (substring string start end)
  (let ((n (%- end start)))
	(let ((r (make-string n)))
	  (let loop ((i 0))
		(if (%eq? i n)
			r
			(loop (begin
					(%string-set! r i (%string-ref string (%+ i start)))
					(%+ i 1))))))))

(define (list->string l)
  (let ((n (length l)))
	(let ((s (make-string n)))
	  (let loop ((i 0)
				 (l l))
		(cond ((null? l) s)
			  (else
			   (%string-set! s i (%car l))
			   (loop (%+ i 1) (%cdr l))))))))

(define (char->string ch)
  (let ((s (make-string 1)))
	(%string-set! s 0 ch)
	s))

(define (string . chars)
  (list->string chars))

(define (string-ref s i)
  (%string-ref s i))

(define (string-set! s i c)
  (%string-set! s i c))

(define (zero-terminate s)
  (string-append s (char->string (ascii->char 0))))

;;(define (buffer-copy src src-start n dst dst-start)
;;  (let loop ((i 0))
;;	(when (%lt? i n)
;;	 (string-set! dst (%+ dst-start i) (%string-ref src (%+ src-start i)))
;;	 (loop (%+ i 1)))))

(define (buffer-copy src src-start n dst dst-start)
  (%buffer-copy src src-start n dst dst-start))

;; ==================================================
;; characters
;; ==================================================	 

(define (integer->char n)
  (%integer->char n))

(define (char->integer ch)
  (%char->integer ch))

(define (ascii->char n)
  (%integer->char n))

(define (char->ascii ch)
  (%char->integer ch))

(define (char? x)
  (%eq? (%get-typecode x) #x02))

(define char-downcase
  (let ((A (%char->integer #\A))
		(Z (%char->integer #\Z)))
	(lambda (ch)
	  (let ((code (%char->integer ch)))
		(if (and (%ge? code A) (%le? code Z))
			(ascii->char (%+ code 32))
			ch)))))

;; ==================================================
;; vectors
;; ==================================================	 

(define (vector? x)
  (%eq? (%get-typecode x) #x18))

(define (vector-ref v i)
  (%tuple-ref v i))

(define (vector-set! v i d)
  (%tuple-set! v i d))

(define (vector-length v)
  (%tuple-length v))

(define (vector-map! p v)
  (let loop ((n (vector-length v)))
	(if (%eq? n 0)
		#f
		(begin
		  (vector-set! v (%- n 1) (p (vector-ref v (%- n 1))))
		  (loop (%- n 1))))))

(define (make-vector n)
  (let ((args (%get-environment))
		(nargs (%- (%tuple-length (%get-environment)) 1))
		(fill #f))
	(if (> nargs 1)
		(set! fill (%tuple-ref args 2)))
	(let ((result (%new-tuple #x18 n)))
	  (vector-map! (lambda (x) fill) result)
	  result)))

(define (vector->list v)
  (let loop ((n (%- (vector-length v) 1))
			 (r '()))
	(if (< n 0)
		r
		(loop (%- n 1) (%cons (vector-ref v n) r)))))

(define (list->vector l)
  (list->tuple l #x18))

(define (vector . l)
  (list->vector l))

;; ==================================================
;; boolean
;; ==================================================	 

(define (boolean? x)
  (%eq? (%get-typecode x) #x06))

(define (not x)
  (%eq? x #f))

;; ==================================================
;; numbers
;; ==================================================	 

(define (+ a b) (%+ a b))
(define (- a b) (%- a b))
(define (* a b) (%* a b))
(define (/ a b) (%/ a b))

(define (= a b) (%eq? a b))
(define (> a b) (%gt? a b))
(define (< a b) (%lt? a b))
(define (>= a b) (%ge? a b))
(define (<= a b) (%le? a b))

(define (zero? x) (= x 0))
(define (negative? x) (< x 0))

(define (quotient a b)
  (%/ a b))

(define (remainder a b)
  (%% a b))
		 
(define (even? x)
  (zero? (remainder x 2)))

(define (odd? x)
  (not (even? x)))

(define (abs x)
  (if (negative? x) (%- 0 x) x))

(define (modulo a b)
  (let ((r (remainder a b)))
	(if (= r 0)
		0
		(if (< a 0)
			(if (< b 0) r (%+ r b))
			(if (< b 0) (%+ r b) r)))))

(define (expt n e)
  (let loop ((i e)
			 (a n))
	(if (eq? e 0)
		a
		(loop (%- i 1) (* a n)))))

(define (number? x)
  (%eq? (%get-typecode x) #x00))

(define (integer? x)
  (%eq? (%get-typecode x) #x00))

(define (number->string n . base)
  (let ((table "0123456789abcdefghijklmnopqrstuvwxyz"))
	(if (null? base)
		(set! base 10)
		(set! base (car base)))
	(if (zero? n)
		"0"
		(let loop ((x (abs n))
				   (r '()))
		  (if (zero? x)
			  (list->string
			   (if (negative? n) (%cons #\- r) r))
			  (loop (quotient x base)
					(%cons (%string-ref table (remainder x base)) r)))))))

;; ==================================================
;; control flow
;; ==================================================

(define (procedure? p)
  (%eq? (%get-typecode p) #xc0))

(define (apply proc args)
  ;; leave room for env link for <extend-env>
  (%apply1 proc (list->tuple (%cons 0 args) #x1c)))

(define (map1 p l)
  (let loop ((l l) (r '()))
	(if (pair? l)
		(loop (%cdr l) (%cons (p (%car l)) r))
		(reverse r)
		)))

(define (map* p l)
  (let loop ((l l) (r '()))
	(if (null? (%car l))
		(reverse r)
		(loop (map1 cdr l)
			  (%cons (apply p (map1 car l)) r))
		)))

;; varargs original
(define (map p . l)
  (if (null? (cdr l))
	  (map1 p (car l))
	  (map* p l)))

 ;; really for-each1
(define (for-each p l)
  (let loop ((l l))
	(if (null? l)
		0
		(begin (p (%car l))
			   (loop (%cdr l))))))

(define (call/cc p)
  (let ((k (%getcc)))
	(p (lambda (result)
		 (%putcc k result)))))

(define (compile-to-closure exp env inline-primitives)
  (let ((tc (transform exp (lambda (x) x) inline-primitives)))
	(let ((cs (compile-cvm tc (car env))))
	  (let ((co (%build-tuple #x28 (list->vector (%cdr cs)) (%car cs))))
		(%build-tuple #x0c co 0 (cdr env))))))

;; taken from repl.scm
(define (eval exp . env)
  (if (null? env)
	  (set! env (interaction-environment))
	  (set! env (car env)))
  (let ((closure (compile-to-closure exp env #f)))
	(closure)))

;; an environment for eval really consists of the
;; the symbol/depth bindings and the actual environment
;; tuple that goes with it.
(define (interaction-environment)
  (cons *cts* *cte*))

;; ==================================================
;; input/output
;; ==================================================	 

;; not really
(define (display x)
  (%print x))

;; to be replaced by write.scm::write
(define (write x)
  (%print x))

(define (newline . port)
  (if (null? port)
	  (set! port stdin)
	  (set! port (car port)))
  ((port 'write-char) (integer->char 10)))

(define (read . port)
  (if (null? port)
	  (set! port stdin)
	  (set! port (car port)))
  (((make-reader port) 'read)))

(define (load filename)
  (let ((port (open-input-file filename)))
	(let ((reader (make-reader port)))
	  (if verbose-interaction (print "; read..."))
	  (let ((forms ((reader 'read-all))))
		((port 'close))
		(if verbose-interaction (print "; compile..."))
		(for-each eval forms);; top-level defines will by dynamic
		))))

(define (read-forms-from-file filename)
  (let ((port (open-input-file filename)))
	(let ((reader (make-reader port)))
	  (let ((forms ((reader 'read-all))))
		((port 'close))
		forms))))

;; ==================================================
;; misc
;; ==================================================	 

(define (check-room nbytes)
  (if (< (%room) nbytes)
	  (>= (%gc) nbytes)
	  #t))
					  
(define (how-many x n)
  (%/ (%- (%+ x n) 1) n))

(define (tuple? x)
  (%eq? (%get-typecode x) #x1c))

;; last result (this should dynamic?)
(define _ #f)
;; compile-time-environment
(define *cte* #f)
;; compile-time symbol-table
(define *cts* #f)

(define (make-environment)
  (let ((table (list)))
	
	(define (new-cell sym)
	  (let ((cell (%cons sym (%undefined))))
		(set! table (%cons cell table))
		cell))

	(define (dynref sym)
	  (let ((probe (%assq sym table)))
		(if probe
			(%cdr probe)
			(error "undefined dynamic symbol" sym))))

	(define (dynset! sym val)
	  (let ((probe (%assq sym table)))
		(let ((cell (if probe probe (new-cell sym))))
		  ;; (set-cdr! cell val)
		  (%tuple-set! cell 1 val) ;; until we get macros
		  )))

	(%cons dynref dynset!)))

;; this is a generic object-dispatcher, it builds a closure
;; that will map symbols to 'methods':
;; (dispatcher one two)
;; =>
;; (let ((table `((one . ,one)
;;                (two . ,two))))
;;   (lambda (msg)
;;     (let ((probe (%assq msg table)))
;;       (if probe
;;           (%cdr probe)
;;           (error "no such method")))))

(define-macro (dispatcher . methods)
  (let ((pairs (list 'quasiquote (map (lambda (name) (cons name (list 'unquote name))) methods)))
		(table (gensym))
		(msg (gensym))
		(probe (gensym)))
	`(let ((,table ,pairs))
	   (lambda (,msg)
		 (let ((,probe (%assq ,msg ,table)))
		   (if ,probe
			   (%cdr ,probe)
			   (error "no such method")))))))

(define (initialize-runtime)

  (%write-string "initializing runtime...") (%write-char (integer->char 10))

  ;; setup

  ;; compile-time-environment
  (set! *cte* (%tuple-ref (%get-environment) 0))
  (set! symbol-table (list))
  (set! assoc-string (assoc-pred string=?))
  (set! member (member-pred equal?))
  (set! memv (member-pred eqv?))

  (set! the-environment (make-environment))
  ;; these are referenced by compiled code
  (set! ~~dynref  (%car the-environment))
  (set! ~~dynset! (%cdr the-environment))

  ;; KLUDGE
  ;; Bootstrap puts a vector of symbol objects where we can get at it.
  (let ((top-env (%tuple-ref (%tuple-ref (%get-environment) 0) 0)))
	(let ((symbols (%tuple-ref top-env 0)))
	  (vector-map! intern-symbol symbols)
	  ;; clear it
	  (%tuple-set! top-env 0 '())
	  ))

  ;; KLUDGE
  ;; Bootstrap puts the compile-time symbols at the end of the code
  ;; object's constant vector.
  
  (let ((cv (%tuple-ref (%tuple-ref eq? 0) 0)))
	(let ((vl (vector-length cv)))
	  (set! *cts* (list '() (vector-ref cv (%- vl 1))))
	  ;; clear it
	  (%tuple-set! cv (%- vl 1) #f)
	  ))
  
  (%write-string "done.") (%write-char (integer->char 10))
  )

