;; -*- Mode: Scheme; tab-width: 4 -*-

(define gensym
  (let ((n 0))
	(lambda ()
	  (set! n (+ n 1))
	  (string->symbol (string-append "g" (number->string n))))))


(define transform

  (let ((inline-substitutions
		 '((cons			%cons)
		   (car				%car)
		   (cdr				%cdr)
		   (eq?				%eq?)
		   (+				%+)
		   (-				%-)
		   (*				%*)
		   (quotient		%/)
		   (=				%eq?)
		   (integer->char	%integer->char)
		   (char->integer	%char->integer)
		   (assq			%assq)
		   (memq			%memq)
		   ))
		(transform-alist '())
		)
	;;
	;; One eyeball on MzScheme's transformers.
	;;
	;; Thanks to Dorai Sitaram and Matthew Flatt for assistance.
	;;
	;; let, let*, letrec, and, or, cond, begin, define, quasiquote,
	;; unquote, unquote-splicing, do, case
	
	(define (raise-syntax-error . l)
	  (error l))

	(define gensym
	  (let ((n 0))
		(lambda ()
		  (set! n (+ n 1))
		  (string->symbol (string-append "g" (number->string n))))))

	;;
	;; The Tricky Bits:
	;; 1) definitions inside bodies.
	;; 2) variable-length argument lists (two forms)
	;;
	;; These nasty warts must be handled within the source tree walker.
	;;

	;; (lambda x <body>)
	;; =>
	;; (let ((x (%cdr (tuple->list (%get_environment))))) <body>)
	;; => ((lambda (x) <body>) (%cdr (tuple->list (%get_environment))))
	;;
	;; (lambda (x y z . l) <body>)
	;; =>
	;; (let ((g23 (%cdr (tuple->list (%get_environment)))))
	;;    (let ((x (%car g23))
	;;          (y (%car (%cdr g23)))
	;;          (z (%car (%cdr (%cdr g23))))
	;;          (l (%cdr (%cdr (%cdr g23)))))
	;;       <body>))
	;; [the %cdr/%car chains are inline primitives, thus the code is
	;;  actually small compared to calling something like 'cadddr']
	;; TODO: we should probably recast this in terms of %tuple-ref.

	(define frob-definition
	  (lambda (e)
		(cond ((pair? (car e))
			   `(,(caar e) (named-lambda ,(caar e) (,@(cdr (car e))) ,@(cdr e))))
			  ((symbol? (car e)) e)
			  (else
			   (raise-syntax-error "bad syntax (define)" e)))))

	(define (frob-varargs e)

	  (define (frob-dotted formals body)
		(let ((sym (gensym)))
		  `(lambda ()
			 (let ((,sym (%cdr (tuple->list (%get-environment)))))
			   (let ,(let loop ((l formals)
								(r '())
								(a sym))
					   (cond ((pair? l)
							  (loop (cdr l)
									(cons `(,(car l) (%car ,a)) r)
									`(%cdr ,a)))
							 (else (reverse (cons `(,l ,a) r)))))
				 ,@body)))))
	
	  (let ((formals (car e))
			(body (cdr e)))
		(cond ((symbol? formals)
			   `(lambda () ((lambda (,formals) ,@body) (%cdr (tuple->list (%get-environment))))))
			  (else (frob-dotted formals body)))))

	(define (collect-matching-forms seq sym)
	  ;; collect forms from <seq> starting with <sym>
	  (let loop ((body seq)
				 (non-matches '())
				 (matches '()))
		(cond ((null? body)
			   (cons non-matches matches))
			  ((and (pair? (car body)) (eq? (caar body) sym))
			   (loop (cdr body) non-matches (cons (car body) matches)))
			  (else
			   (loop (cdr body) (cons (car body) non-matches) matches))
			  )))

	(define (frob-macro m)
	  (let ((formals (cadr m))
			(body (cddr m)))
		(let ((name (car formals))
			  (formals (cdr formals)))
		  `(,name (compile-time-lambda ,formals ,@body)))))

	(define (frob-body e)
	  ;; first, pull out macro definitions
	  (let ((scan (collect-matching-forms e 'define-macro)))
		(let ((forms (car scan))
			  (macrodefs (cdr scan)))
		  ;; then pull out definitions
		  (let ((scan (collect-matching-forms forms 'define)))
			(let ((forms (car scan))
				  (definitions (cdr scan)))
			  (let ((forms
					 (if (null? definitions)
						 forms
						 `((letrec ,(map frob-definition (map cdr definitions)) ,@forms)))))
				(if (not (null? macrodefs))
					`((let-macro
					   ;; (define-macro (<name> <arg0> <arg1> ... ) <body>)
					   ;; => (letmacro ((name (lambda (<arg0> <arg1) <body>))) ...)
					   ,(map frob-macro macrodefs)
					   (begin ,@forms)))
					forms)))))))
			  
	(define (initialize-transforms)

	  (define (undefined)
		(%undefined))

	  (define (define-transform name proc)
		(set! transform-alist (cons (cons name proc) transform-alist)))

	  (define-transform 'and
		(lambda (e)
		  (cond ((null? e) #t)
				((null? (cdr e)) (car e))
				(else `(if ,(car e) (and ,@(cdr e)))))))

	  (define-transform 'or
		(lambda (e)
		  (cond ((null? e) #f)
				((null? (cdr e)) (car e))
				(else
				 (let ((temp (gensym)))
				   `(let ((,temp ,(car e)))
					  (if ,temp
						  ,temp
						  (or ,@(cdr e))))
				   )))))

	  ;; (when <test> <exp1> <exp2> ...)
	  ;; => (if <test>
	  ;;       (begin <exp1> <exp2> ...))

	  (define-transform 'when
		(lambda (e)
		  `(if ,(car e) (begin ,@(cdr e)))))

	  ;; (let <variable0> ((<variable1> <init1>) ...)
	  ;;      <body>)
	  ;; ; =>
	  ;; ((letrec ((<variable0> (lambda (<variable1> ...) <body>)))
	  ;;    <variable0>)
	  ;;  <init1> ...)
	  ;;
	  ;; (let ((<name1> <init1>)
	  ;;       (<name2> <init2>) ...)
	  ;;    <body>)
	  ;;  =>
	  ;; ((lambda (<name1> <name2> ...) <body>) <init1> <init2>)
  
	  (define-transform 'let
		(lambda (e)
		  (if (symbol? (car e))
			  ;; named let
			  (let ((loopvar (car e))
					(vars (map car (cadr e)))
					(inits (map cadr (cadr e)))
					(body (cddr e)))
				`((letrec ((,loopvar (lambda ,vars ,@body)))
					,loopvar)
				  ,@inits))
			  ;; normal let
			  (let ((decls (car e))
					(body (cdr e)))
				(let ((names (map car decls))
					  (inits (map cadr decls)))
				  `((lambda ,names ,@body) ,@inits))))))


	  ;; (let* ((<name1> <init1>)
	  ;;        (<name2> <init2>)
	  ;;        ...)
	  ;;   <body>)
	  ;; =>
	  ;; (let ((<name1> <init1>))
	  ;;   (let* ((<name2> <init2>))
	  ;;      <body>))

	  (define-transform 'let*
		(lambda (e)
		  (let ((decls (car e)))
			(if (null? (cdr decls))
				;; one binding
				(cons 'let e)
				;; two+ bindings
				`(let (,(car decls))
				   (let* (,@(cdr decls)) ,@(cdr e)))))))

	  ;; (letrec ((<name1> <init1)
	  ;;          (<name2> <init2) ... )
	  ;;   <body>)
	  ;; =>
	  ;; (let ((<name1> #f)
	  ;;       (<name2> #f) ... )
	  ;;   (set! <name1> <init1>)
	  ;;   (set! <name2> <init2>)
	  ;;   <body>)
  
	  (define-transform 'letrec
		(lambda (e)
		  (let ((decls (car e))
				(body (cdr e)))
			(let ((empty-decls (map (lambda (d) `(,(car d) #f)) decls))
				  (assigns (map (lambda (d) `(set! ,(car d) ,(cadr d))) decls)))
			  `(let (,@empty-decls)
				 ,@assigns
				 ,@body)))))


	  ;; top-level define
	  (define-transform 'define
		(lambda (e)
		  (let ((fe (frob-definition e)))
			`(~~dynset! (quote ,(car fe)) ,@(cdr fe)))))

	  ;;(define-transform 'define
	  ;;  (lambda (e)
	  ;;	(cons 'set! (frob-definition e))))
  
	  (define-transform 'cond
		(lambda (form)
		  (let loop ((tests form))
			(if (null? tests)
				'(%undefined)
				(if (not (pair? tests))
					(raise-syntax-error "bad syntax (body must contain a list of pairs)" tests)
					(let ((line (car tests))
						  (rest (cdr tests)))
					  (if (not (pair? line))
						  (raise-syntax-error "bad syntax (clause is not a test-value pair)" line)
						  (let* ((test (car line))
								 (value (cdr line))
								 (else? (eq? test 'else)))
							(if (and else? (pair? rest))
								(raise-syntax-error "bad syntax (`else' clause must be last)" line))
							(if (and (pair? value)
									 (eq? '=> (car value)))
								(if (and (pair? (cdr value))
										 (null? (cddr value)))
									(let ((test (if else?
													#t 
													test))
										  (gen (gensym)))
									  `(let ((,gen ,test))
										 (if ,gen
											 (,(cadr value) ,gen)
											 (loop rest))))
									(raise-syntax-error "bad syntax (bad clause form with =>)" line))
								(if else?
									(cons 'begin value)
									(if (null? value)
										(let ((gen (gensym)))
										  `(let ((,gen ,test))
											 (if ,gen ,gen ,(loop rest))))
										(list 'if test (cons 'begin value) (loop rest)))))))))))))

	  ;; taken from JAR's simple-macro package

	  (define-transform 'quasiquote

        (lambda (x)

          (define (expand-quasiquote x level)
            (descend-quasiquote x level finalize-quasiquote))

          (define (finalize-quasiquote mode arg)
            (cond ((eq? mode 'quote) (list 'quote arg))
                  ((eq? mode 'unquote) arg)
                  ((eq? mode 'unquote-splicing)
                   (syntax-error ",@ in invalid context" arg))
                  (else `(,mode ,@arg))))

          (define (descend-quasiquote x level return)
            (cond ((vector? x)
                   (descend-quasiquote-vector x level return))
                  ((not (pair? x))
                   (return 'quote x))
                  ((interesting-to-quasiquote? x 'quasiquote)
                   (descend-quasiquote-pair x (+ level 1) return))
                  ((interesting-to-quasiquote? x 'unquote)
                   (cond ((= level 0)
                          (return 'unquote (cadr x)))
                         (else
                          (descend-quasiquote-pair x (- level 1) return))))
                  ((interesting-to-quasiquote? x 'unquote-splicing)
                   (cond ((= level 0)
                          (return 'unquote-splicing (cadr x)))
                         (else
                          (descend-quasiquote-pair x (- level 1) return))))
                  (else
                   (descend-quasiquote-pair x level return))))

          (define (descend-quasiquote-pair x level return)
            (descend-quasiquote (car x) level
              (lambda (car-mode car-arg)
                (descend-quasiquote (cdr x) level
                  (lambda (cdr-mode cdr-arg)
                    (cond ((and (eq? car-mode 'quote) (eq? cdr-mode 'quote))
                           (return 'quote x))
                          ((eq? car-mode 'unquote-splicing)
                           ;; (,@mumble ...)
                           (cond ((and (eq? cdr-mode 'quote) (null? cdr-arg))
                                  (return 'unquote
                                          car-arg))
                                 (else
                                  (return 'append
                                          (list car-arg (finalize-quasiquote
                                                           cdr-mode cdr-arg))))))
                          (else
                           (return 'cons
                                   (list (finalize-quasiquote car-mode car-arg)
                                         (finalize-quasiquote cdr-mode cdr-arg))))))))))

          (define (descend-quasiquote-vector x level return)
            (descend-quasiquote (vector->list x) level
              (lambda (mode arg)
                (case mode
                  ((quote) (return 'quote x))
                  (else (return 'list->vector
                                (list (finalize-quasiquote mode arg))))))))

          (define (interesting-to-quasiquote? x marker)
            (and (pair? x) (eq? (car x) marker)))

          (expand-quasiquote (car x) 0)))

	  (define-transform 'case
		(lambda (form)
		  (let ((v (car form))
				(tests-in (cdr form))
				(gen (gensym)))
			`(let ((,gen ,v))
			   (cond
				,@(let loop ((tests tests-in))
					(cond
					 ((null? tests) '())
					 ((pair? tests)
					  (let ((test-value (car tests))
							(rest (cdr tests)))
						(cons
						 (if (or (not (pair? test-value))
								 (not (pair? (cdr test-value))))
							 (raise-syntax-error "bad syntax (clause is not a test-value pair)" test-value)
							 (let ((test (car test-value))
								   (value (cdr test-value)))
							   (if (and (pair? value) (list? value))
								   #t
								   (raise-syntax-error "bad syntax (improper clause body)" value))
							   (cons
								(cond
								 ((eq? test 'else)
								  (if (null? rest)
									  'else
									  (raise-syntax-error "bad syntax (`else' clause must be last)" test-value)))
								 (else
								  (cons 
								   'or
								   (let loop ((test test))
									 (cond
									  ((null? test) '())
									  ((pair? test)
									   (let ((t (list 'quote (car test))))
										 (cons `(eqv? ,gen ,t)
											   (loop (cdr test)))))
									  (else (loop (cons test '()))))))))
								`((begin ,@value)))))
						 (loop rest))))
					 (else (raise-syntax-error "bad syntax (body must contain a list of pairs)" tests)))))))))

	  (define-transform 'do
		(lambda (form)
		  (let ((bindings (car form))
				(test-and-result (cadr form))
				(body (cddr form)))
			(let ((code (cons 'do form)))
			  (if (list? bindings)
				  (undefined)
				  (raise-syntax-error 'do "bad syntax (bad binding sequence)" code))
			  (for-each (lambda (binding)
						  (if (and (list? binding)
								   (let ((l (length binding)))
									 (or (= l 2) (= l 3))))
							  (undefined)
							  (raise-syntax-error 'do "bad syntax (bad binding)" code)))
						bindings)
			  (if (list? test-and-result) (undefined)
				  (raise-syntax-error 'do "bad syntax (bad test and result)" code))
			  (if (list? body) (undefined)
				  (raise-syntax-error 'do "bad syntax (improper body)" code))
			  (let ((vars (map (lambda (binding)
								 (let ((var (car binding)))
								   (if (symbol? var)
									   var
									   (raise-syntax-error 'do "bad syntax (variable must be an identifier)" code))))
							   bindings))
					(inits (map cadr bindings))
					(steps (map (lambda (binding)
								  (cond
								   ((pair? (cddr binding))
									(caddr binding))
								   (else (car binding))))
								bindings))
					(test (if (pair? test-and-result)
							  (car test-and-result)
							  (raise-syntax-error 'do "bad syntax (test/result part)" code)))
					(results (if (null? (cdr test-and-result))
								 (list '(undefined))
								 (cdr test-and-result)))
					(loop (gensym)))
				`(let ,loop ,(map (lambda (var init)
									(list var init))
								  vars inits)
					  (if ,test
						  (begin
							,@results)
						  (begin
							,@body
							(,loop ,@steps)))))))))

	  )

	(initialize-transforms)

	(lambda (e p inline-primitives)

	  (let ((substitutions (if inline-primitives
							   inline-substitutions
							   '())))

		(define (varargs? l)
		  (or (symbol? l)
			  (let loop ((l l))
				(cond ((null? l) #f)
					  ((pair? l) (loop (cdr l)))
					  (else #t)))))

		(define (transform* e)
		  (p
		   (if (pair? e)
			   (let ((front (car e)))
				 (case front
				   ;; conditional
				   ((if)
					(if (= (length e) 3)
						`(if ,(transform* (cadr e)) ,(transform* (caddr e)) #f);; #spam
						`(if ,(transform* (cadr e)) ,(transform* (caddr e)) ,(transform* (cadddr e)))))
				   ;; assignment
				   ((set!) `(set! ,(cadr e) ,(transform* (caddr e))))
				   ;; procedure
				   ((named-lambda)
					(if (varargs? (caddr e))
						(transform* (frob-varargs (cddr e)))
						`(named-lambda ,(cadr e) ,(caddr e) ,@(map transform* (frob-body (cdddr e))))))
				   ((lambda)
					(if (varargs? (cadr e))
						(transform* (frob-varargs (cdr e)))
						`(named-lambda #f ,(cadr e) ,@(map transform* (frob-body (cddr e))))))
				   ;; compile-time-lambda
				   ;; this is a lambda that will be eval'd at compile-time.
				   ;; leave it alone.
				   ((compile-time-lambda)
					e)
				   ;; sequence
				   ((begin)
					(if (= (length e) 2)
						(transform* (cadr e))
						(cons 'begin (map transform* (cdr e)))))
				   ;; literal data
				   ((quote)
					e)
				   (else
					(if (symbol? front)
						;; macro
						(let ((macro-probe (assq front transform-alist)))
						  (if macro-probe
							  (transform* ((cdr macro-probe) (cdr e)))
							  (let ((subst-probe (assq front substitutions)))
								(if subst-probe 
									(cons (cadr subst-probe) (map transform* (cdr e)))
									(map transform* e)))))
						;; application
						(map transform* e)))))
			   e)))
		(transform* e))))
	)
